import pandas as pd
import numpy as np
from ..logging import Logging


class TableHorizontal(Logging):


    def __init__(self, *, obj_table):
        super().__init__()
        self.obj_table = obj_table



    def generate_df_horizontal(self):

        obj_table = self.obj_table
        lst_idx_qre = list()
        count = 0
        last_lvl = 0

        for key, val in obj_table.tbl_header.items():

            if obj_table.df_horizontal.empty:
                obj_table.df_horizontal = self.header_lst_qre_to_dataframe(hd_grp=key, lst_tbl_val=val)
            else:
                obj_table.df_horizontal = pd.concat([obj_table.df_horizontal, self.header_lst_qre_to_dataframe(hd_grp=key, lst_tbl_val=val)], axis=0, ignore_index=True)

            # generate idx_qre------------------------------------------------------------------------------------------
            step = 0
            for i in val[-1]:
                step += len(i['cats'].keys())


            for v in range(len(lst_idx_qre), obj_table.df_horizontal.index[-1], step):

                lst_idx_qre.extend([count] * step)

                count += len(val[-1])

            last_lvl = len(val) - 1
            # End of generate idx_qre-----------------------------------------------------------------------------------

        obj_table.df_horizontal['query_combine'] = obj_table.df_horizontal.filter(regex='^query_lvl_[0-9]{1,2}$').agg('&'.join, axis=1)
        obj_table.df_horizontal['idx_qre'] = obj_table.df_horizontal[f'qre_grp_lvl_{last_lvl}'] + lst_idx_qre


        # Add sig test header dataframe (Not yet)


        return self.obj_table



    def header_lst_qre_to_dataframe(self, *, hd_grp: str, lst_tbl_val: list) -> pd.DataFrame:

        df_comb_hd_data = pd.DataFrame()

        self.print(f"Header group name: {hd_grp}", self.clr_cyan_light)

        lst_cols = ['hd_grp', 'hd_lvl', 'qre_grp', 'qre', 'cols', 'type', 'lbl', 'cat_code', 'cat_lbl', 'query']

        for hd_lvl, lst_qre in enumerate(lst_tbl_val):

            self.print(f"Header group level: {hd_lvl}", self.clr_cyan)

            lst_data = self.header_lst_qre_to_dataframe_data(lst_qre=lst_qre, hd_grp=hd_grp, hd_lvl=hd_lvl)

            df_hd_data = pd.DataFrame(columns=lst_cols, data=lst_data)

            df_hd_data.columns = df_hd_data.columns + f'_lvl_{hd_lvl}'

            match hd_lvl:
                case 0:
                    df_comb_hd_data = df_hd_data
                case _:
                    df_comb_hd_data = self.combine_df_header_by_level(df_a=df_comb_hd_data, df_b=df_hd_data)


        return df_comb_hd_data



    def header_lst_qre_to_dataframe_data(self, *, lst_qre: list, hd_grp: str, hd_lvl: int) -> list[list]:

        lst_data = list()

        for i, qre in enumerate(lst_qre):

            self.print(f"Header group question: {i} - {qre}", self.clr_magenta if i % 2 == 0 else self.clr_magenta_light)

            for cat_key, cat_val in qre['cats'].items():

                str_query = str()
                lst_col = list()

                match qre['type']:

                    case 'QUERY':
                        str_query = f"({cat_key})"

                    case 'MA' | 'SA':
                        str_query = '(' + '|'.join([f"{col}=={int(cat_key)}" if int(cat_key) > 0 else f"{col}>{int(cat_key)}" for col in qre['lst_col']]) + ')'
                        lst_col = qre['lst_col']

                    case _:
                        self.print(f"{qre} has invalid type >>> terminated", self.clr_err)

                lst_row = [hd_grp, hd_lvl, i, qre['qre_name'], lst_col, qre['type'], qre['qre_lbl'], cat_key, cat_val, str_query]
                lst_data.append(lst_row)

        return lst_data



    @staticmethod
    def combine_df_header_by_level(df_a: pd.DataFrame, df_b: pd.DataFrame) -> pd.DataFrame:

        # self.print(f"Run 'combine_df_header_by_level'")

        df_a_repeat = pd.DataFrame(np.repeat(df_a.values, df_b.shape[0], axis=0), columns=df_a.columns)
        df_b_repeat = pd.DataFrame(np.repeat(df_b.values, df_a.shape[0], axis=0), columns=df_b.columns)

        df_b_repeat['idx1'] = df_b_repeat.index % df_a.shape[0]
        df_b_repeat['idx2'] = df_b_repeat.index
        df_b_repeat = df_b_repeat.sort_values(by=['idx1', 'idx2']).reset_index(drop=True).drop(columns=['idx1', 'idx2'])

        df_combine = pd.concat([df_a_repeat, df_b_repeat], axis=1)

        return df_combine


















import pandas as pd
import numpy as np
import time
import datetime
import math
import multiprocessing
from scipy import stats
from .logging import Logging



class DataTableGeneratorV2(Logging):

    def __init__(self, *, df_data: pd.DataFrame, df_info: pd.DataFrame, xlsx_name: str, is_md: bool = False, dict_df_tables: dict = None):

        super().__init__()

        pd.set_option('future.no_silent_downcasting', True)

        self.is_md = is_md

        self.df_data, self.df_info = self.convert_md_to_mc(df_data=df_data, df_info=df_info) if is_md else df_data, df_info

        self.file_name = xlsx_name.rsplit('/', 1)[-1] if '/' in xlsx_name else xlsx_name

        self.dict_df_tables = dict_df_tables if dict_df_tables else dict()

        # Unnetted value label------------------------------------------------------------------------------------------
        self.df_info['val_lbl_unnetted'] = self.df_info['val_lbl']

        df_val_lbl = df_info[['var_name', 'val_lbl']].astype(str).query("val_lbl.str.contains('net_code')")

        a = str()
        b = str()
        for idx in df_val_lbl.index:

            if a != str(self.df_info.at[idx, 'val_lbl_unnetted']):
                a = str(self.df_info.at[idx, 'val_lbl_unnetted'])
                b = self.unnetted_qre_val(eval(df_val_lbl.at[idx, 'val_lbl']))

            self.df_info.at[idx, 'val_lbl_unnetted'] = b

        self.print('Unnetted value label - Completed')


        # Check matching of data và defines-----------------------------------------------------------------------------
        tpl_err = self.valcheck_strange_values(df_data=self.df_data, df_info=self.df_info)

        if not tpl_err[0]:
            self.print(tpl_err[1], self.clr_err)
            exit()
        else:
            self.print(tpl_err[1])


        # Remove duplicate value of MA qres------------------------------------------------------------------------------
        self.df_data = self.valcheck_remove_duplicate_ma_vars_values(df_data=self.df_data, df_info=self.df_info)



        # Check file permission-----------------------------------------------------------------------------------------
        try:
            check_perm = open(xlsx_name)
            check_perm.close()

        except PermissionError:
            self.print(f'Permission Error when access file: {xlsx_name}, Processing terminated.', self.clr_err)
            exit()
        except FileNotFoundError:
            pass

        # --------------------------------------------------------------------------------------------------------------



    def unnetted_qre_val(self, dict_netted) -> dict:
        dict_unnetted = dict()

        if 'net_code' not in dict_netted.keys():
            return dict_netted

        for key, val in dict_netted.items():

            if 'net_code' in key:
                val_lbl_lv1 = dict_netted['net_code']

                for net_key, net_val in val_lbl_lv1.items():

                    if isinstance(net_val, str):
                        dict_unnetted.update({str(net_key): net_val})
                    else:
                        self.print(f"Unnetted {net_key}", self.clr_magenta)
                        dict_unnetted.update(net_val)

            else:
                dict_unnetted.update({str(key): val})

        return dict_unnetted



    @staticmethod
    def convert_md_to_mc(*, df_data: pd.DataFrame, df_info: pd.DataFrame) -> (pd.DataFrame, pd.DataFrame):

        df_info_qre_ma_all_1st_col = df_info.query("var_type.isin(['MA', 'MA_mtr']) & var_name.str.contains('[A-Za-z]+_1$')").copy()

        for idx in df_info_qre_ma_all_1st_col.index:
            qre = df_info_qre_ma_all_1st_col.at[idx, 'var_name'].rsplit('_', 1)[0]

            df_info_ma = df_info.query(f"var_name.str.contains('^{qre}_[0-9]+$')").copy()

            dict_val_lbl = dict()
            for idx_ma in df_info_ma.index:
                str_qre, str_val = str(df_info_ma.at[idx_ma, 'var_name']).rsplit('_', 1)
                dict_val_lbl[str_val] = df_info_ma.at[idx_ma, 'val_lbl']['1']

                df_data[df_info_ma.at[idx_ma, 'var_name']] = df_data[df_info_ma.at[idx_ma, 'var_name']].replace({1: int(str_val)})

            df_info.loc[df_info_ma.index,  ['val_lbl']] = [dict_val_lbl]

            lst_ma_col_name = df_info_ma['var_name'].values.tolist()

            df_data[lst_ma_col_name[0]] = df_data[lst_ma_col_name].values.tolist()

            for idx_dt in df_data.index:
                arr = np.sort(df_data.at[idx_dt, lst_ma_col_name[0]], axis=None)
                df_data.at[idx_dt, lst_ma_col_name[0]] = arr

            df_data[lst_ma_col_name] = pd.DataFrame(df_data[lst_ma_col_name[0]].tolist(), index=df_data.index)

        return df_data, df_info



    @staticmethod
    def valcheck_strange_values(*, df_data: pd.DataFrame, df_info: pd.DataFrame) -> (bool, str):

        df_info = df_info.set_index('var_name')
        df_info = df_info.loc[df_info.eval("~var_type.isin(['FT', 'FT_mtr', 'NUM']) | var_name == 'ID'"), :].drop(columns=['var_lbl', 'var_type', 'val_lbl'])

        for idx in df_info.index:
            if idx == 'ID':
                continue

            df_info.at[idx, 'val_lbl_unnetted'] = {int(k): np.nan for k in df_info.at[idx, 'val_lbl_unnetted'].keys()}

        dict_replace = df_info.to_dict()['val_lbl_unnetted']

        df_data = df_data.loc[:, df_info.index].replace(dict_replace).dropna(how='all').dropna(axis=1, how='all')
        df_data = pd.DataFrame(df_data)


        df_data = df_data.reset_index(drop=True if 'ID' in df_data.columns else False)


        df_data = pd.melt(df_data, id_vars=df_data.columns[0], value_vars=df_data.columns[1:]).dropna()

        if not df_data.empty:
            return False, f'Please check values not in codelist:\n{df_data.to_string()}'

        return True, "Check strange values: df_data vs df_info -> Completed"



    def valcheck_remove_duplicate_ma_vars_values(self, df_data: pd.DataFrame, df_info: pd.DataFrame) -> pd.DataFrame:

        # # # REMOVING DUPLICATED VALUES--------------------------------------------------------------------------------
        str_query = "var_name.str.contains(r'^\\w+_1$') & var_type.str.contains('MA')"
        df_info_ma = df_info.query(str_query)

        def remove_dup(row: pd.Series):
            row_idx = row.index.values.tolist()
            lst_val = row.drop_duplicates(keep='first').values.tolist()
            return lst_val + ([np.nan] * (len(row_idx) - len(lst_val)))

        for qre_ma in df_info_ma['var_name'].values.tolist():
            prefix, suffix = qre_ma.rsplit('_', 1)

            self.print(f"Remove duplicate MA qres' values - {prefix}", end='')

            cols = df_info.loc[df_info.eval(f"var_name.str.contains('^{prefix}_[0-9]{{1,2}}$')"), 'var_name'].values.tolist()
            df_data[cols] = df_data[cols].apply(remove_dup, axis=1, result_type='expand')

            print(end='\r')


        self.print("Remove duplicate MA qres' values - Completed")

        return df_data



    def run_tables(self, *, lst_group_tables: list):

        st = time.time()

        lst_run_table = list(self.dict_df_tables.keys())

        for item in lst_group_tables:

            if not item['tables_to_run']:
                self.print('No selected tables. Processing terminated!!!', self.clr_err)
                exit()

            dict_tables = {k: v for k, v in item['tables_format'].items() if k in item['tables_to_run']}

            for tbl_key, tbl_val in dict_tables.items():
                if tbl_val.get('weight_var') and tbl_val.get('sig_test_info').get('sig_type'):
                    self.print(f'Cannot run table "{tbl_key}" with significant test and weighting at the same time. Processing terminated!!!', self.clr_err)
                    exit()

            lst_run_table.extend(item['tables_to_run'])

            self.run_tables_by_item(dict_tables=dict_tables, lst_run_table=lst_run_table)

        self.print(f"All table generation have completed in {datetime.timedelta(seconds=time.time() - st)}", self.clr_succ)



    def run_tables_by_item(self, *, dict_tables: dict, lst_run_table: list):

        is_multiple_process = False
        results = list()

        match is_multiple_process:

            case True:

                # MULTIPLE PROCESSING-----------------------------------------------------------------------------------
                num_cores = multiprocessing.cpu_count()
                num_cores_processing = num_cores - 1
                self.print(f'Number of CPU cores to be used for multiple processing: {num_cores_processing} / {num_cores}', self.clr_warn)

                pool = multiprocessing.Pool(processes=num_cores_processing)

                # Map tasks to the worker function
                results = pool.map(self.run_standard_table_sig, list(dict_tables.values()))

                # Close and join the pool
                pool.close()
                pool.join()
                # End MULTIPLE PROCESSING-------------------------------------------------------------------------------

            case False:

                for item in list(dict_tables.values()):
                    results.append(self.run_standard_table_sig(item))


        # self.dict_df_tables = ({arr[0]: arr[-1] for arr in results})
        self.dict_df_tables.update({arr[0]: arr[-1] for arr in results})
        self.dict_df_tables = dict(sorted(self.dict_df_tables.items(), key=lambda iitem: lst_run_table.index(iitem[0])))

        with pd.ExcelWriter(self.file_name, engine='xlsxwriter') as writer:
            pd.DataFrame(columns=['#', 'Content'], data=[]).to_excel(writer, sheet_name='Content', index=False)

            for k, v in self.dict_df_tables.items():
                v.to_excel(writer, sheet_name=k, index=False)
                self.print(['Save table', k, '- Completed'], [None, self.clr_blue, None], sep=' | ')



    def convert_table_header_to_list(self, lst_header_qres: list) -> list:

        df_info = self.df_info.copy()


        def get_cats_by_qre_name(qre_name: str) -> dict:
            str_qre_name = f'{qre_name[1:]}_1' if '$' in qre_name else qre_name
            return df_info.loc[df_info.eval(f"var_name == '{str_qre_name}'"), 'val_lbl'].values[0]


        def group_sig_table_header_query(dict_qre: dict, str_cat_or_query) -> str:

            if '$' in dict_qre['qre_name']:

                if 'RANK' in str(dict_qre['qre_name']).upper():
                    lst_qre_ma_name = df_info.query(f"var_name.str.contains('^{dict_qre['qre_name'][1:]}[0-9]+$')")['var_name'].values.tolist()
                else:
                    lst_qre_ma_name = df_info.query(f"var_name.str.contains('^{dict_qre['qre_name'][1:]}_[0-9]+$')")['var_name'].values.tolist()

                if str(str_cat_or_query).upper() == 'TOTAL':
                    return f"({' | '.join([f'{i} > 0' for i in lst_qre_ma_name])})"

                return f"({' | '.join([f'{i} == {str_cat_or_query}' for i in lst_qre_ma_name])})"


            if str(str_cat_or_query).upper() == 'TOTAL':
                return f"{dict_qre['qre_name']} > 0"

            if '@' in dict_qre['qre_name']:
                return str_cat_or_query

            return f"{dict_qre['qre_name']} == {str_cat_or_query}"



        lst_group_header = list()

        while len(lst_header_qres) != 5:
            lst_header_qres.append([])

        lvl1, lvl2, lvl3, lvl4, lvl5 = lst_header_qres

        for a in lvl1:

            if not a['cats']:
                a['cats'] = get_cats_by_qre_name(a['qre_name'])

            if not lvl2:
                dict_grp_hd = dict()
                dict_idx = {list(a['cats'].keys())[i]: i for i in range(len(a['cats'].keys()))}

                for a_k, a_v in a['cats'].items():
                    # qa = f"{a['qre_name']} > 0" if str(a_k).upper() == 'TOTAL' else (a_k if '@' in a['qre_name'] else f"{a['qre_name']} == {a_k}")
                    qa = group_sig_table_header_query(a, a_k)

                    dict_grp_hd.update({
                        dict_idx[a_k]: {
                            "lbl": f"{a['qre_lbl']}@{a_v}",
                            "query": f"{qa}"
                        }
                    })

                lst_group_header.append(dict_grp_hd)
                continue

            for a_k, a_v in a['cats'].items():
                # qa = f"{a['qre_name']} > 0" if str(a_k).upper() == 'TOTAL' else (a_k if '@' in a['qre_name'] else f"{a['qre_name']} == {a_k}")
                qa = group_sig_table_header_query(a, a_k)

                for b in lvl2:

                    if not b['cats']:
                        b['cats'] = get_cats_by_qre_name(b['qre_name'])

                    if not lvl3:
                        dict_grp_hd = dict()
                        dict_idx = {list(b['cats'].keys())[i]: i for i in range(len(b['cats'].keys()))}

                        for b_k, b_v in b['cats'].items():
                            # qb = f"{b['qre_name']} > 0" if str(b_k).upper() == 'TOTAL' else f"{b['qre_name']} == {b_k}"
                            qb = group_sig_table_header_query(b, b_k)

                            dict_grp_hd.update({
                                dict_idx[b_k]: {
                                    "lbl": f"{a['qre_lbl']}@{a_v}@{b['qre_lbl']}@{b_v}",
                                    "query": f"{qa} & {qb}"
                                }
                            })

                        lst_group_header.append(dict_grp_hd)
                        continue

                    for b_k, b_v in b['cats'].items():
                        # qb = f"{b['qre_name']} > 0" if str(b_k).upper() == 'TOTAL' else f"{b['qre_name']} == {b_k}"
                        qb = group_sig_table_header_query(b, b_k)

                        for c in lvl3:

                            if not c['cats']:
                                c['cats'] = get_cats_by_qre_name(c['qre_name'])

                            if not lvl4:
                                dict_grp_hd = dict()
                                dict_idx = {list(c['cats'].keys())[i]: i for i in range(len(c['cats'].keys()))}

                                for c_k, c_v in c['cats'].items():
                                    # qc = f"{c['qre_name']} > 0" if str(c_k).upper() == 'TOTAL' else f"{c['qre_name']} == {c_k}"
                                    qc = group_sig_table_header_query(c, c_k)

                                    dict_grp_hd.update({
                                        dict_idx[c_k]: {
                                            "lbl": f"{a['qre_lbl']}@{a_v}@{b['qre_lbl']}@{b_v}@{c['qre_lbl']}@{c_v}",
                                            "query": f"{qa} & {qb} & {qc}"
                                        }
                                    })

                                lst_group_header.append(dict_grp_hd)
                                continue

                            for c_k, c_v in c['cats'].items():
                                # qc = f"{c['qre_name']} > 0" if str(c_k).upper() == 'TOTAL' else f"{c['qre_name']} == {c_k}"
                                qc = group_sig_table_header_query(c, c_k)

                                for d in lvl4:

                                    if not d['cats']:
                                        d['cats'] = get_cats_by_qre_name(d['qre_name'])

                                    if not lvl5:
                                        dict_grp_hd = dict()
                                        dict_idx = {list(d['cats'].keys())[i]: i for i in range(len(d['cats'].keys()))}

                                        for d_k, d_v in d['cats'].items():
                                            # qd = f"{d['qre_name']} > 0" if str(d_k).upper() == 'TOTAL' else f"{d['qre_name']} == {d_k}"
                                            qd = group_sig_table_header_query(d, d_k)

                                            dict_grp_hd.update({
                                                dict_idx[d_k]: {
                                                    "lbl": f"{a['qre_lbl']}@{a_v}@{b['qre_lbl']}@{b_v}@{c['qre_lbl']}@{c_v}@{d['qre_lbl']}@{d_v}",
                                                    "query": f"{qa} & {qb} & {qc} & {qd}"
                                                }
                                            })

                                        lst_group_header.append(dict_grp_hd)
                                        continue

                                    for d_k, d_v in d['cats'].items():
                                        # qd = f"{d['qre_name']} > 0" if str(d_k).upper() == 'TOTAL' else f"{d['qre_name']} == {d_k}"
                                        qd = group_sig_table_header_query(d, d_k)

                                        for e in lvl5:

                                            if not e['cats']:
                                                e['cats'] = get_cats_by_qre_name(e['qre_name'])

                                            dict_grp_hd = dict()
                                            dict_idx = {list(e['cats'].keys())[i]: i for i in range(len(e['cats'].keys()))}

                                            for e_k, e_v in e['cats'].items():
                                                # qe = f"{e['qre_name']} > 0" if str(e_k).upper() == 'TOTAL' else f"{e['qre_name']} == {e_k}"
                                                qe = group_sig_table_header_query(e, e_k)

                                                dict_grp_hd.update({
                                                    dict_idx[e_k]: {
                                                        "lbl": f"{a['qre_lbl']}@{a_v}@{b['qre_lbl']}@{b_v}@{c['qre_lbl']}@{c_v}@{d['qre_lbl']}@{d_v}@{e['qre_lbl']}@{e_v}",
                                                        "query": f"{qa} & {qb} & {qc} & {qd} & {qe}"
                                                    }
                                                })

                                            lst_group_header.append(dict_grp_hd)

        return lst_group_header




    def convert_table_header_to_dataframe(self, header_group: int, lst_header_qres: list) -> pd.DataFrame:

        # PENDING: NOT YET DONE

        df_info = self.df_info.copy()

        def get_cats_by_qre_name(qre_name: str) -> dict:
            str_qre_name = f'{qre_name[1:]}_1' if '$' in qre_name else qre_name
            return df_info.loc[df_info.eval(f"var_name == '{str_qre_name}'"), 'val_lbl'].values[0]



        def get_header_query_by_cats(qre_name: str, qre_lbl: str, tpl_cat: tuple) -> list:

            str_header_label = f"{qre_lbl}@{tpl_cat[1]}"

            if '$' in qre_name:

                if 'RANK' in str(qre_name).upper():
                    str_query = f"var_name.str.contains('^{qre_name[1:]}[0-9]+$')"
                else:
                    str_query = f"var_name.str.contains('^{qre_name[1:]}_[0-9]+$')"

                lst_qre_ma_name = df_info.query(str_query)['var_name'].values.tolist()

                if tpl_cat[0].upper() == 'TOTAL':
                    return [str_header_label, f"({' | '.join([f'{i} > 0' for i in lst_qre_ma_name])})"]

                return [str_header_label, f"({' | '.join([f'{i} == {tpl_cat[0]}' for i in lst_qre_ma_name])})"]



            if tpl_cat[0].upper() == 'TOTAL':
                return [str_header_label, f"({qre_name} > 0)"]

            if '@' in qre_name:
                return [str_header_label, tpl_cat[0]]

            return [str_header_label, f"({qre_name} == {tpl_cat[0]})"]



        def combine_df_header_by_level(df_a: pd.DataFrame, df_b: pd.DataFrame) -> pd.DataFrame:

            df_a_repeat = pd.DataFrame(np.repeat(df_a.values, df_b.shape[0], axis=0), columns=df_a.columns)
            df_b_repeat = pd.DataFrame(np.repeat(df_b.values, df_a.shape[0], axis=0), columns=df_b.columns)

            df_b_repeat['idx1'] = df_b_repeat.index % df_a.shape[0]
            df_b_repeat['idx2'] = df_b_repeat.index
            df_b_repeat = df_b_repeat.sort_values(by=['idx1', 'idx2']).reset_index(drop=True).drop(columns=['idx1', 'idx2'])

            df_combine = pd.concat([df_a_repeat, df_b_repeat], axis=1)

            return df_combine


        df_header = pd.DataFrame()

        for i, item in enumerate(lst_header_qres):

            df_qre = pd.DataFrame()

            for j, qre in enumerate(item):

                if len(qre.get('cats')) == 0:
                    qre['cats'] = get_cats_by_qre_name(qre['qre_name'])

                qre.update({'header_info': []})

                for cat in qre['cats'].items():
                    qre['header_info'].append([j] + get_header_query_by_cats(qre['qre_name'], qre['qre_lbl'], cat))

                df_qre_temp = pd.DataFrame(columns=[f'group_{i}', f'label_{i}', f'query_{i}'], data=qre['header_info'])
                df_qre = pd.concat([df_qre, df_qre_temp], axis=0)


            if df_header.empty:
                df_header = df_qre
            else:
                df_header = combine_df_header_by_level(df_header, df_qre)


        df_header['header_group'] = header_group

        lst_col = list(df_header.columns)
        lst_col = lst_col[-1:] + lst_col[:-1]

        df_header = df_header[lst_col]

        if len(lst_header_qres) > 1:

            # ADD GROUP_[last level]------------------------------------------------------------------------------------
            last_lvl = len(lst_header_qres) - 1
            lst_level_count_cats = list()

            for qre in lst_header_qres[-1]:
                lst_level_count_cats.append(len(qre['cats'].keys()))

            idx_min = 0
            idx_max = sum(lst_level_count_cats) - 1

            for i in range(0, int(df_header.shape[0] / sum(lst_level_count_cats)) * 2, 2):
                df_header.loc[idx_min:idx_max, f'group_{last_lvl}'] = df_header.loc[idx_min:idx_max, f'group_{last_lvl}'] + i
                idx_min = idx_max + 1
                idx_max = idx_max + sum(lst_level_count_cats)

            df_header[f'group_{last_lvl}'] = df_header[f'group_{last_lvl}'].astype(int)


            # ADD GROUP_[before last level]-----------------------------------------------------------------------------

            for ilvl in range(len(lst_header_qres) - 1):

                if ilvl > 0:
                    df_header[f'group_{ilvl}'] = df_header[[f'label_{ilvl - 1}', f'label_{ilvl}']].agg('|'.join, axis=1)
                else:
                    df_header[f'group_{ilvl}'] = df_header[f'label_{ilvl}']


                lst_val_label = df_header[f'group_{ilvl}'].drop_duplicates(keep='first').values.tolist()
                df_header[f'group_{ilvl}'] = df_header[f'group_{ilvl}'].replace({v: i for i, v in enumerate(lst_val_label)})


        # COMBINE QUERY COLUMNS-----------------------------------------------------------------------------------------
        df_header.loc[:, ['query_combined']] = df_header.filter(regex='^query_[0-9]$').agg(' & '.join, axis=1)

        return df_header



    def run_standard_table_sig(self, tbl: dict) -> list:

        pd.set_option('future.no_silent_downcasting', True)

        df_tbl = pd.DataFrame()

        # create df_data with tbl_filter in json file
        df_data = self.df_data.query(tbl['tbl_filter']).copy() if tbl.get('tbl_filter') else self.df_data.copy()


        # create df_info with lst_side_qres in json file
        df_info = pd.DataFrame(columns=['var_name', 'var_lbl', 'var_type', 'val_lbl', 'val_lbl_unnetted', 'qre_fil'], data=[])

        for idx, qre in enumerate(tbl['lst_side_qres']):

            if '$' in qre['qre_name']:

                if '_RANK' in str(qre['qre_name']).upper():
                    lst_qre_col = self.df_info.loc[self.df_info['var_name'].str.contains(f"^{qre['qre_name'][1:]}[0-9]+$"), 'var_name'].values.tolist()
                else:
                    lst_qre_col = self.df_info.loc[self.df_info['var_name'].str.contains(f"^{qre['qre_name'][1:]}_[0-9]+$"), 'var_name'].values.tolist()

                var_name = qre['qre_name'].replace('$', '')

            elif '#combine' in qre['qre_name']:
                var_name, str_comb = qre['qre_name'].split('#combine')
                lst_qre_col = str_comb.replace('(', '').replace(')', '').split(',')
            else:
                lst_qre_col = [qre['qre_name']]
                var_name = qre['qre_name']

            # NEW-------------------------------------------------------------------------------------------------------
            df_qre_info = self.df_info.query(f"var_name.isin({lst_qre_col})").copy()
            df_qre_info = df_qre_info.reset_index(drop=True)

            if df_qre_info.empty:
                self.print(f"\n\tQuestion(s) is not found: {qre['qre_name']}\n\tProcess terminated.", self.clr_err)
                exit()

            dict_row = {
                'var_index': int(idx),
                'var_name': var_name,
                'var_lbl': qre['qre_lbl'].replace('{lbl}', df_qre_info.at[0, 'var_lbl']) if qre.get('qre_lbl') else df_qre_info.at[0, 'var_lbl'],
                'var_type': 'MA_comb' if '#combine' in qre['qre_name'] else ('MA_Rank' if '$' in qre['qre_name'] and '_RANK' in str(qre['qre_name']).upper() else df_qre_info.at[0, 'var_type']),
                'val_lbl': qre['cats'] if qre.get('cats') else df_qre_info.at[0, 'val_lbl'],
                'val_lbl_unnetted': df_qre_info.at[0, 'val_lbl_unnetted'],
                'qre_fil': qre['qre_filter'] if qre.get('qre_filter') else "",
                'lst_qre_col': lst_qre_col,
                'mean': qre['mean'] if qre.get('mean') else {},
                'sort': qre['sort'] if qre.get('sort') else "",
                'calculate': qre['calculate'] if qre.get('calculate') else {},
                'friedman': qre['friedman'] if qre.get('friedman') else {},
                'weight_var': tbl['weight_var'] if tbl.get('weight_var') else "",
            }

            df_info = pd.concat([df_info, pd.DataFrame(columns=list(dict_row.keys()), data=[list(dict_row.values())])], axis=0, ignore_index=True)
            df_info = df_info.set_index('var_index', drop=False)




            # ----------------------------------------------------------------------------------------------------------


        if tbl.get('lst_header_qres'):

            # PENDING
            # df_group_header = self.convert_table_header_to_dataframe(0, tbl['lst_header_qres'])  # test

            # Maximum 5 levels of header
            lst_group_header = self.convert_table_header_to_list(tbl['lst_header_qres'])

        else:
            # TO DO: Run multiple header with same level
            lst_group_header = list()

            df_group_header = pd.DataFrame()  # test

            lvl_hd = -1
            for int_header_group, (key_hd, val_hd) in enumerate(tbl['dict_header_qres'].items()):

                if lvl_hd == -1:
                    lvl_hd = len(val_hd)
                else:
                    if lvl_hd != len(val_hd):
                        self.print(f"Header don't have the same level: {tbl['dict_header_qres']}", self.clr_err)
                        exit()

                # PENDING
                # df_group_header = pd.concat([df_group_header, self.convert_table_header_to_dataframe(int_header_group, val_hd)], axis='rows')  # test

                # Maximum 5 levels for each header
                lst_group_header.extend(self.convert_table_header_to_list(val_hd))


        for grp_hd in lst_group_header:

            tbl_info_sig = {
                'tbl_name': tbl['tbl_name'],
                'is_count': tbl['is_count'],
                'is_pct_sign': tbl['is_pct_sign'],
                'sig_test_info': tbl['sig_test_info'],
                'dict_grp_header': grp_hd,
                'weight_var': tbl.get('weight_var') if tbl.get('weight_var') else ''
            }

            df_temp = self.run_standard_header_sig(df_data, df_info, tbl_info_sig=tbl_info_sig)

            if df_tbl.empty:
                df_tbl = df_temp
            else:
                lst_col_temp_to_add = list(df_temp.columns)[6:]
                df_tbl = pd.concat([df_tbl, df_temp[lst_col_temp_to_add]], axis=1)


        # drop row which have all value is nan
        df_tbl = df_tbl.dropna(how='all')

        # Drop rows in qre oe that have all columns are 0
        if tbl['is_hide_oe_zero_cats']:

            df_sum_oe_val = df_tbl.query("qre_name.str.contains('_OE') & qre_type == 'MA'").copy()

            if not df_sum_oe_val.empty:
                fil_col = list(df_sum_oe_val.columns)
                df_sum_oe_val = df_sum_oe_val.loc[:, fil_col[5:]]
                df_sum_oe_val = df_sum_oe_val.replace({'': np.nan, 0: np.nan})

                # df_sum_oe_val = df_sum_oe_val.astype(float)
                # df_sum_oe_val['sum_val'] = df_sum_oe_val.sum(axis=1, skipna=True, numeric_only=True)

                df_sum_oe_val['sum_val'] = df_sum_oe_val.count(axis=1, numeric_only=True)

                df_sum_oe_val = df_sum_oe_val.query('sum_val == 0')

                df_tbl = df_tbl.drop(df_sum_oe_val.index)


        # # Drop columns which all value equal 0
        # if tbl['is_hide_zero_cols']:
        #
        #     start_idx = df_tbl.query(f"cat_val == 'base'").index.tolist()[0]
        #     lst_val_col = [v for i, v in enumerate(df_tbl.columns.tolist()[5:]) if i % 2 == 0]
        #
        #     df_fil = df_tbl.query("index >= @start_idx")[lst_val_col].copy()
        #     df_fil = df_fil.replace({0: np.nan}).dropna(axis='columns', how='all')
        #
        #     lst_keep_col = list()
        #     for i in df_fil.columns.tolist():
        #         lst_keep_col.extend([i, i.replace('@val@', '@sig@')])
        #
        #     df_tbl = df_tbl[df_tbl.columns.tolist()[:5] + lst_keep_col]
        #
        #     # df_tbl.to_excel('df_tbl_review.xlsx')


        # Add number to header for formatting
        here = 1



        # Add number to header for formatting


        return [tbl['tbl_name'], df_tbl]




    @staticmethod
    def add_base_to_tbl_sig(df_qre: pd.DataFrame, qre_info: dict, dict_header_col_name: dict, lst_sig_pair: list, weight_var: str = '') -> pd.DataFrame:

        lst_tbl_row_data = list()
        lst_ignore_col = list()

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            for idx_item, item in enumerate(sig_pair):

                if item in lst_ignore_col:
                    continue

                df_filter = dict_header_col_name[item]['df_data'].copy()

                if len(weight_var) > 0:
                    lst_qre_col = qre_info['lst_qre_col']
                    df_fil_base = df_filter[lst_qre_col].dropna(how='all')
                else:
                    df_fil_base = df_filter.dropna(how='all')

                df_filter = df_filter.loc[df_fil_base.index, :]

                if df_filter.empty:
                    num_base = 0
                else:
                    num_base = df_filter.shape[0]

                if len(weight_var) > 0:
                    mean_weight = df_filter.loc[:, weight_var].mean()
                    num_base *= mean_weight

                if len(lst_tbl_row_data) == 0:
                    str_qre_name = qre_info['qre_name']
                    str_lbl = 'Weighted Base' if len(weight_var) > 0 else 'Base'
                    lst_tbl_row_data = [qre_info['qre_index'], str_qre_name, qre_info['qre_lbl'], qre_info['qre_type'], 'base', str_lbl, num_base, np.nan]
                else:
                    lst_tbl_row_data.extend([num_base, np.nan])

                lst_ignore_col.append(item)

        df_qre.loc[len(df_qre)] = lst_tbl_row_data
        return df_qre



    def add_sa_qre_val_to_tbl_sig(self, df_qre: pd.DataFrame, qre_info: dict, dict_header_col_name: dict,
                                  lst_sig_pair: list, sig_type: str, lst_sig_lvl: list,
                                  cat: str, lbl: str, lst_sub_cat: list | None, weight_var: str = '') -> pd.DataFrame:

        qre_index = qre_info['qre_index']
        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']
        qre_val = qre_info['qre_val_unnetted']
        is_count = qre_info['is_count']
        val_pct = qre_info['val_pct']

        dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
        dict_new_row.update({
            'qre_index': qre_index,
            'qre_name': qre_name,
            'qre_lbl': qre_lbl,
            'qre_type': qre_type,
            'cat_val': cat,
            'cat_lbl': lbl,
        })

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        lst_ran_col = list()

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):

                if not sig_type and item in lst_ran_col:
                    continue

                if item not in lst_ran_col:
                    lst_ran_col.append(item)

                # df_filter = dict_header_col_name[item]['df_data'].loc[:, [qre_name]].copy()
                df_filter = dict_header_col_name[item]['df_data'].copy()

                # # HERE: OPTIMIZING
                # df_des = df_filter.describe()
                # df_count = df_filter.value_counts()
                # df_pct = df_filter.value_counts(normalize=True)

                if df_filter.empty:
                    continue

                if lst_sub_cat:
                    dict_re_qre_val = {int(k): 1 if k in lst_sub_cat else 0 for k, v in qre_val.items()}
                else:
                    dict_re_qre_val = {int(k): 1 if k == cat else 0 for k, v in qre_val.items()}

                df_filter = df_filter.replace(dict_re_qre_val)

                dict_pair_to_sig.update({item: df_filter})

                # UPDATE FOR WEIGHTED TABLE
                if len(weight_var) > 0:

                    df_temp_for_weight = df_filter.loc[df_filter.eval(f"~{qre_name}.isnull()"), qre_name].copy()

                    if df_temp_for_weight.empty:
                        num_val = np.nan
                    else:
                        if is_count:
                            num_val = np.dot(df_temp_for_weight, dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var])
                        else:
                            num_val = np.average(df_temp_for_weight, weights=dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var]) * val_pct

                else:
                    num_val = (df_filter[qre_name] == 1).sum() if is_count else df_filter[qre_name].mean() * val_pct

                val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                if sig_type and lst_sig_lvl:
                    num_val_old = df_qre.loc[df_qre['cat_val'] == cat, [val_col_name]].values[0, 0]

                    if pd.isnull(num_val_old):
                        df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:

                    df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]




            if sig_type and lst_sig_lvl and len(weight_var) == 0:
                df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)




        return df_qre



    def add_sa_qre_mean_to_tbl_sig(self, df_qre: pd.DataFrame, qre_info: dict, dict_header_col_name: dict,
                                   lst_sig_pair: list, sig_type: str, lst_sig_lvl: list, mean_factor: dict,
                                   is_mean: bool, weight_var: str = '', is_friedman_sig: bool = False) -> pd.DataFrame:

        qre_index = qre_info['qre_index']
        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']
        # qre_val = qre_info['qre_val']
        qre_val = qre_info['qre_val_unnetted']

        if is_mean:
            dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
            dict_new_row.update({
                'qre_index': qre_index,
                'qre_name': qre_name,
                'qre_lbl': qre_lbl,
                'qre_type': qre_type,
                'cat_val': 'mean',
                'cat_lbl': 'Mean',
            })

        elif is_friedman_sig:

            dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
            dict_new_row.update({
                'qre_index': qre_index,
                'qre_name': qre_name,
                'qre_lbl': qre_lbl,
                'qre_type': qre_type,
                'cat_val': 'friedman_pval',
                'cat_lbl': 'Friedman P-value',
            })

        else:
            dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
            dict_new_row.update({
                'qre_index': qre_index,
                'qre_name': qre_name.replace('_Mean', '_Std'),
                'qre_lbl': qre_lbl,
                'qre_type': qre_type,
                'cat_val': 'std',
                'cat_lbl': 'Std',
            })

        org_qre_name = qre_name.replace('_Mean', '')

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        lst_ran_col = list()

        dict_sig_friedman = dict()

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):

                if not sig_type and item in lst_ran_col:
                    continue

                if item not in lst_ran_col:
                    lst_ran_col.append(item)

                # df_filter = dict_header_col_name[item]['df_data'].loc[:, [org_qre_name]].copy()
                df_filter = dict_header_col_name[item]['df_data'].copy()


                if df_filter.empty:
                    continue

                if mean_factor:
                    df_filter = df_filter.replace({org_qre_name: mean_factor})

                if -999 not in qre_val.keys():

                    try:

                        df_filter = df_filter.replace(qre_val)

                    except TypeError as err:
                        raise err


                dict_pair_to_sig.update({item: df_filter})

                if is_friedman_sig:
                    if item in dict_sig_friedman.keys():
                        continue

                    dict_sig_friedman.update({item: df_filter})

                if is_mean:

                    # UPDATE FOR WEIGHTED TABLE
                    if len(weight_var) > 0:

                        df_temp_for_weight = df_filter.loc[df_filter.eval(f"~{org_qre_name}.isnull()"), org_qre_name].copy()

                        if df_temp_for_weight.empty:
                            num_val = np.nan
                        else:
                            num_val = np.average(df_temp_for_weight, weights=dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var])
                    else:
                        num_val = df_filter[org_qre_name].mean()


                    val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                    if sig_type and lst_sig_lvl:

                        num_val_old = df_qre.loc[df_qre['cat_val'] == 'mean', [val_col_name]].values[0, 0]

                        if pd.isnull(num_val_old):
                            df_qre.loc[df_qre['cat_val'] == 'mean', [val_col_name, sig_col_name]] = [num_val, np.nan]

                    else:
                        df_qre.loc[df_qre['cat_val'] == 'mean', [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:

                    # UPDATE FOR WEIGHTED TABLE
                    if len(weight_var) > 0:

                        df_temp_for_weight = df_filter.loc[df_filter.eval(f"~{org_qre_name}.isnull()"), org_qre_name].copy()

                        if df_temp_for_weight.empty:
                            num_val_std = np.nan
                        else:
                            df_weight = dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var].copy()
                            num_mean = np.average(df_temp_for_weight, weights=df_weight)
                            num_variance = np.average((df_temp_for_weight - num_mean) ** 2, weights=df_weight)
                            num_val_std = math.sqrt(num_variance)
                    else:
                        num_val_std = df_filter[org_qre_name].std()

                    val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                    df_qre.loc[df_qre['cat_val'] == 'std', [val_col_name, sig_col_name]] = [num_val_std, np.nan]

            if len(weight_var) > 0 or (not is_mean and not is_friedman_sig):
                continue

            if is_mean:
                if sig_type and lst_sig_lvl:
                    df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)


        if is_friedman_sig:

            lst_val_col_name = list()
            for k in dict_sig_friedman.keys():
                lst_val_col_name.append(dict_header_col_name[k]['val_col'])

            if len(dict_sig_friedman.keys()) < 3:
                df_qre.loc[df_qre['cat_val'] == 'friedman_pval', lst_val_col_name] = ['NAN'] * len(lst_val_col_name)

            else:
                df_friedman_sig = pd.DataFrame()

                for k, df in dict_sig_friedman.items():
                    df_frm_temp = df.reset_index(drop=True)
                    df_friedman_sig = pd.concat([df_friedman_sig, df_frm_temp], axis=1, ignore_index=True)

                arr_friedman_sig = df_friedman_sig.to_numpy()
                arr_friedman_sig = arr_friedman_sig.transpose()
                frm = stats.friedmanchisquare(*arr_friedman_sig)

                df_qre.loc[df_qre['cat_val'] == 'friedman_pval', lst_val_col_name] = [frm.pvalue] * len(lst_val_col_name)


        return df_qre



    def add_sa_qre_group_to_tbl_sig(self, df_qre: pd.DataFrame, qre_info: dict, dict_header_col_name: dict,
                                    lst_sig_pair: list, sig_type: str, lst_sig_lvl: list, cat: str, lbl: str,
                                    weight_var: str = '') -> pd.DataFrame:


        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']
        qre_val = qre_info['qre_val']
        is_count = qre_info['is_count']
        val_pct = qre_info['val_pct']

        dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
        dict_new_row.update({
            'qre_name': qre_name,
            'qre_lbl': qre_lbl,
            'qre_type': qre_type,
            'cat_val': cat,
            'cat_lbl': lbl,
        })

        org_qre_name = qre_name.replace('_Group', '')

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        lst_ran_col = list()

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):

                if not sig_type and item in lst_ran_col:
                    continue

                if item not in lst_ran_col:
                    lst_ran_col.append(item)

                df_filter = dict_header_col_name[item]['df_data'].loc[:, [org_qre_name]].copy()

                if df_filter.empty:
                    continue

                df_filter = df_filter.replace(qre_val['recode'])

                dict_re_qre_val = {int(k): 1 if int(k) == int(cat) else 0 for k, v in qre_val['cats'].items()}
                df_filter = df_filter.replace(dict_re_qre_val)

                dict_pair_to_sig.update({item: df_filter})

                # UPDATE FOR WEIGHTED TABLE
                if len(weight_var) > 0:

                    df_temp_for_weight = df_filter.loc[df_filter.eval(f"~{org_qre_name}.isnull()"), org_qre_name].copy()

                    if df_temp_for_weight.empty:
                        num_val = np.nan
                    else:
                        if is_count:
                            num_val = np.dot(df_temp_for_weight, dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var])
                        else:
                            num_val = np.average(df_temp_for_weight, weights=dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var]) * val_pct

                else:
                    num_val = (df_filter[org_qre_name] == 1).sum() if is_count else df_filter[org_qre_name].mean() * val_pct

                val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                if sig_type and lst_sig_lvl:

                    num_val_old = df_qre.loc[df_qre['cat_val'] == cat, [val_col_name]].values[0, 0]

                    if pd.isnull(num_val_old):
                        df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:

                    df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]


            if sig_type and lst_sig_lvl and len(weight_var) == 0:
                df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)

        return df_qre



    @staticmethod
    def add_sa_qre_cal_to_tbl_sig(df_qre: pd.DataFrame, qre_info: dict, dict_cal: dict) -> pd.DataFrame:

        df_qre['idx_by_cat_lbl'] = df_qre['cat_lbl']
        df_qre = df_qre.set_index('idx_by_cat_lbl')

        for key, val in dict_cal.items():

            dict_new_row = {col: '' if '@sig@' in col else dict_cal.get('syntax') for col in df_qre.columns}
            dict_new_row.update({
                'qre_index': qre_info['qre_index'],
                'qre_name': qre_info['qre_name'],
                'qre_lbl': qre_info['qre_lbl'],
                'qre_type': qre_info['qre_type'],
                'cat_val': f'calculate|{key}',
                'cat_lbl': key,
            })

            df_temp = pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])

            # df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

            df_temp['idx_by_cat_lbl'] = df_temp['cat_lbl']
            df_temp = df_temp.set_index('idx_by_cat_lbl')

            for col in df_temp.columns:
                if '@val@' not in col:
                    continue

                str_syntax = val.replace('[', f"df_qre.loc['").replace(']', f"', '{col}']")

                df_temp.loc[key, col] = eval(str_syntax)

            # HERE: Find solution for warning
            # a = 0

            df_qre = pd.concat([df_qre, df_temp], axis=0)

            # a = 1


        df_qre = df_qre.reset_index(drop=True)

        return df_qre



    def add_num_qre_to_tbl_sig(self, df_qre: pd.DataFrame, qre_info: dict, dict_header_col_name: dict, cal_act: str,
                               lst_sig_pair: list, sig_type: str, lst_sig_lvl: list, weight_var: str = '') -> pd.DataFrame:

        # Add option: std, quantile 25/50/75, min, max
        dict_cal_act = {
            'mean': 'Mean',
            'std': 'Std',
            'min': 'Minimum',
            'max': 'Maximum',
            '25%': 'Quantile 25%',
            '50%': 'Quantile 50%',
            '75%': 'Quantile 75%',
        }

        qre_index = qre_info['qre_index']
        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']

        dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
        dict_new_row.update({
            'qre_index': qre_index,
            'qre_name': qre_name,
            'qre_lbl': qre_lbl,
            'qre_type': qre_type,
            'cat_val': cal_act,
            'cat_lbl': dict_cal_act.get(cal_act),
        })

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        lst_ran_col = list()

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):

                if not sig_type and item in lst_ran_col:
                    continue

                if item not in lst_ran_col:
                    lst_ran_col.append(item)

                # df_filter = dict_header_col_name[item]['df_data'].loc[:, [qre_name]].copy()
                df_filter = dict_header_col_name[item]['df_data'].copy()

                if df_filter.empty:
                    continue

                dict_pair_to_sig.update({item: df_filter})

                # UPDATE FOR WEIGHTED TABLE
                if len(weight_var) > 0:

                    df_temp_for_weight = df_filter.loc[df_filter.eval(f"~{qre_name}.isnull()"), qre_name].copy()

                    if df_temp_for_weight.empty:
                        num_val = np.nan
                    else:
                        df_weight = dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var].copy()

                        if cal_act in ['std']:
                            num_mean = np.average(df_temp_for_weight, weights=df_weight)
                            num_variance = np.average((df_temp_for_weight - num_mean) ** 2, weights=df_weight)
                            num_val = math.sqrt(num_variance)
                        elif cal_act in ['min', 'max']:
                            num_val = df_filter[qre_name].describe().loc[cal_act]

                        elif cal_act in ['25%', '50%', '75%']:

                            def weighted_percentile(data, percents, weights=None):
                                """ percents in units of 1%
                                weights specifies the frequency (count) of data.
                                """
                                if weights is None:
                                    return np.percentile(data, percents)
                                ind = np.argsort(data)
                                d = data[ind]
                                w = weights[ind]
                                p = 1. * w.cumsum() / w.sum() * 100
                                y = np.interp(percents, p, d)
                                return y

                            num_val = weighted_percentile(np.array(df_temp_for_weight), int(cal_act.replace('%', '')), np.array(df_weight))

                        else:
                            num_val = np.average(df_temp_for_weight, weights=dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var])

                else:
                    num_val = df_filter[qre_name].describe().loc[cal_act]

                val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                if sig_type and lst_sig_lvl:

                    num_val_old = df_qre.loc[df_qre['cat_val'] == cal_act, [val_col_name]].values[0, 0]

                    if pd.isnull(num_val_old):
                        df_qre.loc[df_qre['cat_val'] == cal_act, [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:
                    df_qre.loc[df_qre['cat_val'] == cal_act, [val_col_name, sig_col_name]] = [num_val, np.nan]

            if sig_type and lst_sig_lvl and len(weight_var) == 0 and cal_act == 'mean':
                df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)

        return df_qre



    def add_ma_qre_val_to_tbl_sig(self, df_qre: pd.DataFrame, qre_info: dict, dict_header_col_name: dict,
                                  lst_sig_pair: list, sig_type: str, lst_sig_lvl: list, cat: str, lbl: str,
                                  lst_sub_cat: list | None, weight_var: str = '') -> pd.DataFrame:

        if lst_sub_cat is None:
            lst_sub_cat = []

        qre_index = qre_info['qre_index']
        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']
        qre_val = qre_info['qre_val_unnetted']
        is_count = qre_info['is_count']
        val_pct = qre_info['val_pct']
        # df_ma_info = qre_info['df_ma_info']
        lst_qre_col = qre_info['lst_qre_col']

        dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
        dict_new_row.update({
            'qre_index': qre_index,
            'qre_name': qre_name,  # qre_name.rsplit('_', 1)[0],
            'qre_lbl': qre_lbl,
            'qre_type': qre_type,
            'cat_val': cat,
            'cat_lbl': lbl,
        })

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        lst_ran_col = list()

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):

                if not sig_type and item in lst_ran_col:
                    continue

                if item not in lst_ran_col:
                    lst_ran_col.append(item)

                # df_filter = dict_header_col_name[item]['df_data'].loc[:, lst_qre_col].copy()
                df_filter = dict_header_col_name[item]['df_data'].copy()


                if df_filter.empty:
                    continue

                if lst_sub_cat:
                    dict_re_qre_val = {int(k): 1 if k in lst_sub_cat else 0 for k, v in qre_val.items()}

                else:
                    dict_re_qre_val = {int(k): 1 if k == cat else 0 for k, v in qre_val.items()}

                df_filter = df_filter.replace(dict_re_qre_val)

                if len(weight_var) > 0:
                    df_fil_base = df_filter[lst_qre_col].dropna(how='all')

                else:
                    df_fil_base = df_filter.dropna(how='all')

                df_filter.loc[df_fil_base.index, 'ma_val_sum'] = df_filter.loc[df_fil_base.index, lst_qre_col].sum(axis='columns')

                if lst_sub_cat or qre_type == 'MA_comb':
                    df_filter.loc[df_filter['ma_val_sum'] > 1, 'ma_val_sum'] = 1

                dict_pair_to_sig.update({item: df_filter['ma_val_sum']})

                # UPDATE FOR WEIGHTED TABLE
                if len(weight_var) > 0:

                    df_temp_for_weight = df_filter.loc[df_filter.eval(f"~ma_val_sum.isnull()"), 'ma_val_sum'].copy()

                    if df_temp_for_weight.empty:
                        num_val = np.nan
                    else:

                        if is_count:
                            num_val = np.dot(df_temp_for_weight, dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var])
                        else:
                            num_val = np.average(df_temp_for_weight, weights=dict_header_col_name[item]['df_data'].loc[df_temp_for_weight.index, weight_var]) * val_pct

                else:
                    num_val = (df_filter['ma_val_sum'] == 1).sum() if is_count else df_filter['ma_val_sum'].mean() * val_pct



                val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                if sig_type and lst_sig_lvl:
                    num_val_old = df_qre.loc[df_qre['cat_val'] == cat, [val_col_name]].values[0, 0]

                    if pd.isnull(num_val_old):
                        df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:
                    df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]

            if sig_type and lst_sig_lvl and len(weight_var) == 0:
                df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)

        return df_qre



    @staticmethod
    def mark_sig_to_df_qre(df_qre: pd.DataFrame, dict_pair_to_sig: dict, sig_pair: list, dict_header_col_name: dict, sig_type: str, lst_sig_lvl: list) -> pd.DataFrame:

        if not lst_sig_lvl or not sig_type or not dict_pair_to_sig:
            return df_qre

        if sig_pair[0] not in dict_pair_to_sig.keys() or sig_pair[1] not in dict_pair_to_sig.keys():
            return df_qre

        df_left, df_right = dict_pair_to_sig[sig_pair[0]], dict_pair_to_sig[sig_pair[1]]

        if df_left.shape[0] < 30 or df_right.shape[0] < 30:
            return df_qre

        is_df_left_null = df_left.isnull().values.all()
        is_df_right_null = df_right.isnull().values.all()

        if is_df_left_null or is_df_right_null:
            return df_qre

        try:
            if df_left.mean().iloc[0] == 0 or df_right.mean().iloc[0] == 0:
                return df_qre

            if df_left.mean().iloc[0] == 1 and df_right.mean().iloc[0] == 1:
                return df_qre

        except Exception:
            if df_left.mean() == 0 or df_right.mean() == 0:
                return df_qre

            if df_left.mean() == 1 and df_right.mean() == 1:
                return df_qre

        arr_left = df_left.to_numpy(dtype=float)
        arr_right = df_right.to_numpy(dtype=float)

        import warnings
        warnings.filterwarnings("ignore", category=RuntimeWarning)

        match sig_type:
            case 'rel':
                if df_left.shape[0] != df_right.shape[0]:
                    return df_qre

                sigResult = stats.ttest_rel(df_left, df_right)

            case 'ind':
                sigResult = stats.ttest_ind(arr_left, arr_right)

            case _:

                raise ValueError(sig_type)



        if sigResult.pvalue:
            if sigResult.statistic > 0:
                mark_sig_char = sig_pair[1]
                sig_col_name = dict_header_col_name[sig_pair[0]]['sig_col']
            else:
                mark_sig_char = sig_pair[0]
                sig_col_name = dict_header_col_name[sig_pair[1]]['sig_col']

            if sigResult.pvalue <= lst_sig_lvl[0]:
                df_qre.at[df_qre.index[-1], sig_col_name] = str(df_qre.at[df_qre.index[-1], sig_col_name]).replace('nan', '') + mark_sig_char.upper()
            elif len(lst_sig_lvl) >= 2:
                if sigResult.pvalue <= lst_sig_lvl[1]:
                    df_qre.loc[df_qre.index[-1], sig_col_name] = str(df_qre.at[df_qre.index[-1], sig_col_name]).replace('nan', '') + mark_sig_char.lower()

        return df_qre



    def run_standard_header_sig(self, df_data: pd.DataFrame, df_info: pd.DataFrame, tbl_info_sig: dict) -> pd.DataFrame:

        is_count = tbl_info_sig['is_count']
        val_pct = 1 if tbl_info_sig['is_pct_sign'] else 100

        sig_type = tbl_info_sig['sig_test_info']['sig_type']

        lst_sig_lvl_pct = tbl_info_sig['sig_test_info']['lst_sig_lvl']
        lst_sig_lvl = [(100 - a) / 100 for a in lst_sig_lvl_pct]
        lst_sig_lvl.reverse()

        dict_grp_header = tbl_info_sig['dict_grp_header']

        dict_char_sig = {0: 'A', 1: 'B', 2: 'C', 3: 'D', 4: 'E', 5: 'F', 6: 'G', 7: 'H', 8: 'I', 9: 'J', 10: 'K', 11: 'L', 12: 'M', 13: 'N', 14: 'O', 15: 'P', 16: 'Q', 17: 'R', 18: 'S', 19: 'T', 20: 'U', 21: 'V', 22: 'W', 23: 'X', 24: 'Y'}

        # lst_tbl_col = ['qre_name', 'qre_lbl', 'qre_type', 'cat_val', 'cat_lbl']
        dict_tbl_data = {
            'qre_index': list(),
            'qre_name': list(),
            'qre_lbl': list(),
            'qre_type': list(),
            'cat_val': list(),
            'cat_lbl': list(),
        }

        dict_header_col_name_origin = dict()
        # df_data_query = pd.DataFrame()

        for hd_k, hd_v in dict_grp_header.items():
            str_hd_val = f"{hd_v['query']}@{hd_v['lbl']}@val@{dict_char_sig[hd_k]}"
            str_hd_sig = f"{hd_v['query']}@{hd_v['lbl']}@sig@{dict_char_sig[hd_k]}"

            dict_tbl_data.update({
                str_hd_val: str_hd_val.split('@'),
                str_hd_sig: str_hd_sig.split('@'),
            })

            try:
                df_data_query = df_data.query(hd_v['query']).copy()
            except ValueError:
                self.print(f"\tValueError, Cannot process: {hd_v}", self.clr_err)
                exit()
            except pd.errors.UndefinedVariableError:
                self.print(f"\tpandas.errors.UndefinedVariableError, Cannot process: {hd_v}", self.clr_err)
                exit()

            dict_header_col_name_origin.update({
                dict_char_sig[hd_k]: {
                    'val_col': str_hd_val,
                    'sig_col': str_hd_sig,
                    'df_data': df_data_query,
                }
            })

            if len(dict_tbl_data['qre_name']) == 0:
                arr_nan = [np.nan] * len(str_hd_val.split('@'))
                dict_tbl_data.update({
                    'qre_index': arr_nan,
                    'qre_name': arr_nan,
                    'qre_lbl': arr_nan,
                    'qre_type': arr_nan,
                    'cat_val': arr_nan,
                    'cat_lbl': arr_nan,
                })

        if sig_type in ["ind", "rel"]:

            if tbl_info_sig['sig_test_info']['sig_cols']:
                lst_sig_pair = tbl_info_sig['sig_test_info']['sig_cols']
            else:
                lst_sig_char = list(dict_header_col_name_origin.keys())
                lst_sig_pair = list()
                for i in range(len(lst_sig_char)-1):
                    for j in range(i + 1, len(lst_sig_char)):
                        lst_sig_pair.append([lst_sig_char[i], lst_sig_char[j]])

        else:
            lst_sig_char = list(dict_header_col_name_origin.keys())
            lst_sig_pair = list()

            for i in lst_sig_char:
                lst_sig_pair.append([i])


        df_tbl = pd.DataFrame.from_dict(dict_tbl_data)

        df_tbl['qre_lbl'] = df_tbl['qre_lbl'].astype('object')

        lst_tbl_info = [f"Cell content: {'count' if is_count else ('percentage(%)' if tbl_info_sig['is_pct_sign'] else 'percentage')}"]

        if not is_count and sig_type != '':
            lst_tbl_info.extend([
                f"{'Dependent' if sig_type == 'rel' else 'Independent'} Pair T-test at level {' & '.join([f'{i}%' for i in lst_sig_lvl_pct])}",
                f"Columns Tested: {', '.join(['/'.join(i) for i in lst_sig_pair])}",
                f"Uppercase for {lst_sig_lvl_pct[-1]}%, lowercase for {lst_sig_lvl_pct[0]}%" if len(lst_sig_lvl_pct) > 1 else np.nan
            ])

        if len(tbl_info_sig['weight_var']) > 0:
            lst_tbl_info.extend([f"Weighted with: {tbl_info_sig['weight_var']}"])

        df_tbl.loc[1:len(lst_tbl_info), ['qre_lbl']] = lst_tbl_info

        # lst_ft_qre_warn = list()

        for idx in df_info.index:

            qre_idx = int(df_info.at[idx, 'var_index'])
            qre_name = df_info.at[idx, 'var_name']
            qre_lbl = df_info.at[idx, 'var_lbl']
            qre_type = df_info.at[idx, 'var_type']
            qre_val = eval(df_info.at[idx, 'val_lbl']) if isinstance(df_info.at[idx, 'val_lbl'], str) else df_info.at[idx, 'val_lbl']
            qre_val_unnetted = df_info.at[qre_idx, 'val_lbl_unnetted']
            qre_fil = df_info.at[idx, 'qre_fil']
            lst_qre_col = df_info.at[idx, 'lst_qre_col']
            weight_var = df_info.at[idx, 'weight_var']
            lst_qre_col_weight_var = lst_qre_col if len(weight_var) == 0 else lst_qre_col + [weight_var]

            dict_header_col_name = dict()
            for key, val in dict_header_col_name_origin.items():
                dict_header_col_name[key] = val.copy()

                if len(qre_fil) > 0:
                    dict_header_col_name[key]['df_data'] = dict_header_col_name[key]['df_data'].query(qre_fil)

                dict_header_col_name[key]['df_data'] = dict_header_col_name[key]['df_data'][lst_qre_col_weight_var]

            qre_info = {
                'qre_index': qre_idx,
                'qre_name': qre_name,
                'qre_lbl': qre_lbl,
                'qre_type': qre_type,
                'qre_val': qre_val,
                'qre_val_unnetted': qre_val_unnetted,
                'is_count': is_count,
                'val_pct': val_pct,
                'lst_qre_col': lst_qre_col,
            }

            lst_grp_header_item = list()
            for v in dict_grp_header.values():
                lst_grp_header_item.extend(v['lbl'].rsplit('@', 1))

            lst_grp_header_item = list(dict.fromkeys(lst_grp_header_item))

            self.print(
                ['Populate table', tbl_info_sig['tbl_name'], f"{lst_grp_header_item[0]} - [{', '.join(lst_grp_header_item[1:])}]", f"{qre_name} - {qre_type}"],
                [None, self.clr_blue, self.clr_cyan_light, self.clr_magenta_light],
                sep=' | ', end='')





            df_qre = pd.DataFrame(columns=df_tbl.columns, data=[])

            # BASE------------------------------------------------------------------------------------------------------
            df_qre = self.add_base_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, weight_var)
            # END BASE--------------------------------------------------------------------------------------------------

            if qre_type in ['FT', 'FT_mtr']:
                # Not run free text questions
                # lst_ft_qre_warn.extend([f'{qre_name}|{qre_type}'])
                pass

            elif qre_type in ['NUM']:

                if not qre_info.get('qre_val'):
                    qre_info['qre_val'] = {
                        'mean': 'Mean',
                        'std': 'Std',
                        'min': 'Minimum',
                        'max': 'Maximum',
                        '25%': 'Quantile 25%',
                        '50%': 'Quantile 50%',
                        '75%': 'Quantile 75%',
                    }

                for key_num_opt in qre_info['qre_val'].keys():
                    df_qre = self.add_num_qre_to_tbl_sig(df_qre, qre_info, dict_header_col_name, key_num_opt, lst_sig_pair, sig_type, lst_sig_lvl, weight_var)


            elif qre_type in ['SA', 'SA_mtr', 'RANKING']:

                # qre_info['qre_val'] = df_info.at[qre_name, 'val_lbl_unnetted']

                # HERE
                # NEED TO OPTIMIZE

                for cat, lbl in qre_val.items():

                    # ADD-IN Net Code
                    if 'net_code' in str(cat):

                        for net_cat, net_val in qre_val['net_code'].items():

                            if isinstance(net_val, dict):

                                lst_sub_cat = list(net_val.keys())

                                # "900001|combine|POSITIVE (NET)"
                                # list_net_cat = net_cat.split('|')
                                net_cat_val, net_cat_type, net_cat_lbl = net_cat.split('|')

                                df_qre = self.add_sa_qre_val_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, net_cat_val, net_cat_lbl, lst_sub_cat, weight_var)

                                if 'NET' in net_cat_type.upper():
                                    for cat2, lbl2 in net_val.items():
                                        df_qre = self.add_sa_qre_val_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, cat2, f' - {lbl2}', None, weight_var)

                                df_qre.loc[df_qre['cat_val'] == net_cat_val, 'cat_val'] = f'{net_cat_val}|{net_cat_type}'

                            else:
                                df_qre = self.add_sa_qre_val_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, net_cat, net_val, None, weight_var)

                    else:
                        df_qre = self.add_sa_qre_val_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, cat, lbl, None, weight_var)

                mean_factor = df_info.at[idx, 'mean']
                if mean_factor.keys():
                    # Run Mean
                    df_qre = self.add_sa_qre_mean_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, mean_factor, is_mean=True, weight_var=weight_var)

                    # Run Std
                    df_qre = self.add_sa_qre_mean_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, mean_factor, is_mean=False, weight_var=weight_var)

                friedman_factor = df_info.at[idx, 'friedman']
                if friedman_factor.keys():
                    df_qre = self.add_sa_qre_mean_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, mean_factor=friedman_factor, is_mean=False, weight_var=weight_var, is_friedman_sig=True)


                dict_cal = df_info.at[idx, 'calculate']
                if dict_cal:
                    df_qre = self.add_sa_qre_cal_to_tbl_sig(df_qre, qre_info, dict_cal)


            elif qre_type in ['MA', 'MA_mtr', 'MA_comb', 'MA_Rank']:

                # qre_info['qre_val'] = df_info.at[qre_name, 'val_lbl_unnetted']

                for cat, lbl in qre_val.items():

                    # ADD-IN Net Code
                    if 'net_code' in cat:

                        for net_cat, net_val in qre_val['net_code'].items():

                            if isinstance(net_val, dict):

                                lst_sub_cat = list(net_val.keys())

                                # "900001|combine|POSITIVE (NET)"
                                # list_net_cat = net_cat.split('|')
                                net_cat_val, net_cat_type, net_cat_lbl = net_cat.split('|')

                                df_qre = self.add_ma_qre_val_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, net_cat_val, net_cat_lbl, lst_sub_cat, weight_var)

                                if 'NET' in net_cat_type.upper():
                                    for cat2, lbl2 in net_val.items():
                                        df_qre = self.add_ma_qre_val_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, cat2, f' - {lbl2}', None, weight_var)

                                df_qre.loc[df_qre['cat_val'] == net_cat_val, 'cat_val'] = f'{net_cat_val}|{net_cat_type}'

                            else:
                                df_qre = self.add_ma_qre_val_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, net_cat, net_val, None, weight_var)

                    else:
                        df_qre = self.add_ma_qre_val_to_tbl_sig(df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, cat, lbl, None, weight_var)




            # # BUGGING!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # # SORTING---------------------------------------------------------------------------------------------------
            # sort_opt = df_info.at[idx, 'sort']
            #
            # # if sort_opt:
            # #     is_asc = True if sort_opt == 'asc' else False
            # #     base_val = -999_999_999 if sort_opt == 'asc' else 999_999_999
            # #
            # #     df_qre['sort_col'] = df_qre[df_qre.columns.tolist()[5]]
            # #     df_qre.loc[df_qre['cat_val'] == 'base', 'sort_col'] = base_val
            # #
            # #     df_qre.sort_values(by=['sort_col'], ascending=is_asc, inplace=True, ignore_index=True)
            # #     df_qre.drop(columns=['sort_col'], inplace=True)
            #
            # # END SORTING-----------------------------------------------------------------------------------------------


            df_tbl = pd.concat([df_tbl, df_qre], axis=0, ignore_index=True)

            print(end='\r')


        # self.print(f'Cannot create table for free text questions: {lst_ft_qre_warn}', self.clr_warn)

        return df_tbl


    



import pandas as pd
from pydantic import BaseModel, Field, PositiveInt, model_validator
from typing import Optional, Dict, List, Self, Any
from enum import Enum



class QType(str, Enum):
    SA = 'SA'
    MA = 'MA'
    NUM = 'NUM'
    FT = 'FT'
    SA_MTR = 'SA_mtr'
    MA_MTR = 'MA_mtr'



class NumAtt(str, Enum):
    MEAN = 'mean'
    STD = 'std'
    MIN = 'min'
    MAX = 'max'
    _25 = '25%'
    _50 = '50%'
    _75 = '75%'



class CatGroupStt(str, Enum):
    COMBINE = 'combine'
    NET = 'net'



class CatAtt(BaseModel):
    code: str = Field(min_length=1, frozen=True)
    label: str = Field(min_length=1, frozen=True)

    factor: Optional[int | float] = Field(default=None)

    group_stt: Optional[CatGroupStt] = Field(default=None)
    group_code: Optional[object] = Field(default=None)





class Question(BaseModel):

    name: str = Field(min_length=2, frozen=True)
    label: Optional[str] = Field(min_length=2, default=None)
    type: Optional[QType] = Field(default=None)
    cols: Optional[List[str]] = Field(default=None)

    obj_cats: Optional[List[CatAtt]] = Field(default=None)
    obj_num: Optional[List[NumAtt]] = Field(default=None)

    cats: Optional[Dict[str | int, str | dict]] = Field(default=None)
    mean_factor: Optional[Dict[PositiveInt, float]] = Field(default=None)

    filter: Optional[str] = Field(default=None)
    sort: Optional[str] = Field(default=None, pattern=r'acs|des')
    calculation: Optional[Dict[str, str]] = Field(default=None, examples=[{"lbl": "Sum(T2B, B2B)", "syntax": "[T2B] + [B2B]"}])


    def fill_question(self, df_qre: pd.DataFrame) -> Self:

        # LABEL --------------------------------------------------------------------------------------------------------
        if self.label is None:
            self.label = df_qre['var_lbl'].values[0]

        elif '{lbl}' in self.label:
            self.label = self.label.replace('{lbl}', df_qre['var_lbl'].values[0])


        # TYPE ---------------------------------------------------------------------------------------------------------
        self.type = df_qre['var_type'].values[0]

        # COLUMNS ------------------------------------------------------------------------------------------------------
        self.cols = df_qre['var_name'].values.tolist()


        # CATEGORICAL ATTRIBUTES ---------------------------------------------------------------------------------------
        # MEAN FACTOR ---------------------------------------------------------------------------------------

        if self.cats is None:
            self.cats = df_qre['val_lbl'].values[0]

        if (self.type not in [QType.SA, QType.SA_MTR]) and self.mean_factor is not None:
            raise ValueError(f"Attribute 'mean_factor' only use for SA question!!!")

        if self.type in [QType.SA, QType.SA_MTR, QType.MA, QType.MA_MTR]:
            self.obj_cats: List[CatAtt] = self.find_cat(self.cats)


        if self.type != QType.NUM and self.obj_num is not None:
            raise ValueError(f"Attribute 'num_att' only use for NUM question!!!")



        return self





    def find_cat(self, dict_cats: dict) -> List[CatAtt]:

        obj_cats: List[CatAtt] = list()

        for code, label in dict_cats.items():
            ojb_cat = self.find_cat_recur(code, label)

            obj_cats.append(ojb_cat)


        return obj_cats




    def find_cat_recur(self, code: str, label: str | dict) -> CatAtt:

        if isinstance(label, str):

            cat_att = CatAtt(code=code, label=label, factor=self.mean_factor.get(int(code)) if self.mean_factor is not None else None)

        else:

            if code.count('|') != 2:
                raise ValueError(f"{code} has incorrect structure!!!")

            code2, grp_stt, label2 = code.split('|')
            factor = self.mean_factor.get(int(code2)) if self.mean_factor is not None else None
            cat_att = CatAtt(code=code2, label=label2, factor=factor, group_stt=grp_stt, group_code=self.find_cat(label))

        return cat_att








import pandas as pd
import numpy as np
from pydantic import BaseModel, Field, ConfigDict, model_validator
from typing import Self
from ..logging import Logging

log = Logging()


class Data2Run(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    is_validated: bool = False
    is_md: bool = False
    df_data: pd.DataFrame = Field(min_items=1)
    df_info: pd.DataFrame = Field(min_items=1)


    @model_validator(mode='after')
    def valcheck_dfs(self) -> Self:

        if self.is_validated:
            log.print(f"Data2Run is already checked", log.clr_succ)
            return self

        log.print(f"Data2Run validating", log.clr_cyan)

        df_data = self.df_data
        df_info = self.df_info

        # # # COLUMNS CONSISTENCY-----------------------------------------------------------------------------------
        if df_data.shape[1] != df_info.shape[0]:
            raise ValueError("df_data's columns are not match with df_info's rows")

        log.print(" - Columns consistency")


        # # # CHECK & CONVERT 'val_lbl' to dict if it's a string----------------------------------------------------
        if df_info['val_lbl'].dtype in [str, object]:

            def convert_to_dict(row):

                if row == '{}':
                    return {}
                elif isinstance(row, dict):
                    return row
                else:
                    return eval(row)

            df_info['val_lbl'] = df_info['val_lbl'].apply(convert_to_dict)

            log.print(" - Check & Covert 'val_lbl' to dict if it is a string")


        # # # CONVERT TO MC-----------------------------------------------------------------------------------------
        if self.is_md:
            def recode_md_to_mc(row: pd.Series):
                lst_re = [i + 1 for i, v in enumerate(row.values.tolist()) if v == 1]
                return lst_re + ([np.nan] * (len(row.index) - len(lst_re)))

            def create_info_mc(row: pd.Series):
                lst_val = row.values.tolist()
                dict_re = {str(i + 1): v['1'] for i, v in enumerate(lst_val)}
                return [dict_re] * len(lst_val)

            for idx in df_info.query("var_type.isin(['MA', 'MA_mtr']) & var_name.str.contains(r'^\\w+\\d*_1$')").index:
                qre = df_info.at[idx, 'var_name'].rsplit('_', 1)[0]
                fil_idx = df_info.eval(f"var_name.str.contains('^{qre}_[0-9]+$')")
                cols = df_info.loc[fil_idx, 'var_name'].values.tolist()

                df_data[cols] = df_data[cols].apply(recode_md_to_mc, axis=1, result_type='expand')
                df_info.loc[fil_idx, ['val_lbl']] = df_info.loc[fil_idx, ['val_lbl']].apply(create_info_mc, result_type='expand')

            log.print(" - Convert to MC dataframe")


        # # # UNNETTED----------------------------------------------------------------------------------------------
        df_info['val_lbl_str'] = df_info['val_lbl'].astype(str)
        df_info['val_lbl_unnetted'] = df_info['val_lbl']

        for idx in df_info.query("val_lbl_str.str.contains('net_code')").index:

            dict_netted = df_info.at[idx, 'val_lbl_unnetted']
            dict_unnetted = dict()

            for key, val in dict_netted.items():

                if 'net_code' in key:
                    val_lbl_lv1 = dict_netted['net_code']

                    for net_key, net_val in val_lbl_lv1.items():

                        if isinstance(net_val, str):
                            dict_unnetted.update({str(net_key): net_val})
                        else:
                            dict_unnetted.update(net_val)

                else:
                    dict_unnetted.update({str(key): val})

            df_info.at[idx, 'val_lbl_unnetted'] = dict_unnetted

        df_info.drop(columns='val_lbl_str', inplace=True)

        log.print(" - Unnetted: 'val_lbl' in df_info")


        # # # CHECK UNIDENTIFIED VALUES-----------------------------------------------------------------------------
        df_data_check = df_data.copy()
        df_info_check = df_info.copy()

        df_info_check = df_info_check.loc[df_info_check.eval("~var_type.isin(['FT', 'FT_mtr', 'NUM']) | var_name == 'ID'"), :].drop(columns=['var_lbl', 'var_type', 'val_lbl'])
        df_data_check = df_data_check[df_info_check['var_name'].values.tolist()].dropna(axis=1, how='all').dropna(axis=0, how='all')
        df_info_check = df_info_check.set_index('var_name').loc[df_data_check.columns.tolist(), :]

        def convert_val_lbl(row):
            if row[0] != {}:
                row[0] = {int(k): np.nan for k in row[0].keys()}
            return row

        df_info_check = df_info_check.apply(convert_val_lbl, axis=1)
        dict_replace = df_info_check.to_dict()['val_lbl_unnetted']

        df_data_check = df_data_check.replace(dict_replace).dropna(axis=1, how='all')

        cols = df_data_check.columns.tolist()

        if 'ID' in cols:
            cols.remove('ID')

        df_data_check = df_data_check.dropna(subset=cols, how='all', axis=0)

        if not df_data_check.empty:
            df_data_check.reset_index(drop=True if 'ID' in df_data_check.columns else False, inplace=True)
            df_data_check = pd.melt(df_data_check, id_vars=df_data_check.columns[0], value_vars=df_data_check.columns[1:]).dropna()

            raise ValueError(f"Unidentified values are detected\n{df_data_check.to_string()}")

        log.print(" - Check unidentified values")


        # # # REMOVING DUPLICATED VALUES----------------------------------------------------------------------------
        str_query = "var_name.str.contains(r'^\\w+_1$') & var_type.str.contains('MA')"
        df_info_ma = df_info.query(str_query)

        def remove_dup(row: pd.Series):
            row_idx = row.index.values.tolist()
            lst_val = row.drop_duplicates(keep='first').values.tolist()
            return lst_val + ([np.nan] * (len(row_idx) - len(lst_val)))

        for qre_ma in df_info_ma['var_name'].values.tolist():
            prefix, suffix = qre_ma.rsplit('_', 1)
            cols = df_info.loc[df_info.eval(f"var_name.str.contains('^{prefix}_[0-9]{{1,2}}$')"), 'var_name'].values.tolist()
            df_data[cols] = df_data[cols].apply(remove_dup, axis=1, result_type='expand')

        log.print(" - Removing duplicated values")

        return self




import pandas as pd
import numpy as np
from ..logging import Logging





class TabulationVertical(Logging):


    def __init__(self, *, obj_table):
        super().__init__()
        self.obj_table = obj_table



    def generate_df_vertical(self):

        lst_col = ['name', 'label', 'type', 'cat_code', 'cat_label', 'cols', 'factor', 'filter', 'sort', 'calculation']


        for item in self.obj_table.tbl_side:

            print(item.name, item.label, item.type)



        # here:



        a = 1




        return self.obj_table















from inspect import currentframe, getframeinfo
from colorama import Fore, Back
from datetime import datetime
import re
import pathlib
import colorama


class Logging:

    def __init__(self):

        self.max_len = 63

        self.clr_err = Fore.RED
        self.clr_warn = Fore.LIGHTYELLOW_EX
        self.clr_succ = Fore.LIGHTGREEN_EX

        self.clr_blue = Fore.BLUE
        self.clr_cyan = Fore.CYAN
        self.clr_magenta = Fore.MAGENTA

        self.clr_blue_light = Fore.LIGHTBLUE_EX
        self.clr_cyan_light = Fore.LIGHTCYAN_EX
        self.clr_magenta_light = Fore.LIGHTMAGENTA_EX

        self.clr_reset = Fore.RESET



    def print(self, txt: str | list[str], fore_color: None | list[Fore] = None, sep: str = '', end: str = '\n'):

        frameinfo = getframeinfo(currentframe().f_back)
        filename = re.split(r"[/|\\]", frameinfo.filename)[-1]
        linenumber = frameinfo.lineno
        now = datetime.now().strftime("%d-%m-%Y %H:%M:%S")

        str_prefix = f"{Fore.LIGHTBLACK_EX}{now} {self.clr_blue}{filename}:{linenumber}{Fore.RESET}"
        str_prefix += " " * (self.max_len - len(str_prefix))

        lst_txt = [txt] if not isinstance(txt, list) else txt
        lst_fore_color = [fore_color] if not isinstance(fore_color, list) else fore_color
        lst_content = list()

        for t, c in zip(lst_txt, lst_fore_color):
            str_suffix = re.sub('Completed', f'{self.clr_succ}Completed{self.clr_reset}', str(t))
            str_suffix = f"{c if c else ''}{str_suffix}{Fore.RESET}"
            lst_content.extend([str_suffix])


        str_content = f"{str_prefix} {sep.join(lst_content) if len(lst_content) > 1 else lst_content[0]}"

        print(str_content, end=end)







import pandas as pd
import numpy as np
from scipy.stats import kendalltau

from .logging import Logging


class DataProcessing(Logging):

    def __init__(self, *, df_data: pd.DataFrame, df_info: pd.DataFrame):
        super().__init__()
        self.df_data: pd.DataFrame = df_data
        self.df_info: pd.DataFrame = df_info


    @staticmethod
    def str_formater(lst_input: list[str]) -> str:
        return ', '.join(lst_input) if len(lst_input) <= 5 else f"{', '.join(lst_input[:2])},..., {', '.join(lst_input[-2:])}"


    def add_qres(self, dict_add_new_qres: dict, is_add_oe_col: bool = False) -> (pd.DataFrame, pd.DataFrame):
        """
        :param dict_add_new_qres:
            'var_name': ['var_lbl', 'var_type', val_lbl, default value],
            var_name: str
            var_lbl: str
            var_type: str ['SA', 'SA_mtr', 'MA', 'MA_mtr', 'NUM', 'FT']
        :param is_add_oe_col: bool
        :return: df_data, df_info
        """

        info_col_name = ['var_name', 'var_lbl', 'var_type', 'val_lbl']

        lst_data_addin = list()
        lst_info_addin = list()
        lst_colname = list()
        int_max_row = self.df_data.shape[0]

        for key, val in dict_add_new_qres.items():
            self.print(['Add new variables to df_data & df_info', key], [None, self.clr_blue], sep=': ', end='')

            if val[1] in ['MA']:
                qre_ma_name, max_col = str(key).rsplit('|', 1)

                for i in range(1, int(max_col) + 1):

                    lst_info_addin.append([f'{qre_ma_name}_{i}', val[0], val[1], val[2]])

                    if '_OE' not in key or is_add_oe_col is True:
                        lst_colname.append(f'{qre_ma_name}_{i}')
                        lst_data_addin.append([val[-1]] * int_max_row)

            else:

                lst_info_addin.append([key, val[0], val[1], val[2]])

                if '_OE' not in key or is_add_oe_col is True:
                    lst_colname.append(key)
                    lst_data_addin.append([val[-1]] * int_max_row)

            print(end='\r')


        self.df_info = pd.concat([self.df_info, pd.DataFrame(columns=info_col_name, data=lst_info_addin)], axis=0, ignore_index=True)

        if len(lst_colname) > 0:
            self.df_data = pd.concat([self.df_data, pd.DataFrame(columns=lst_colname, data=np.array(lst_data_addin).transpose())], axis=1)

        self.df_data.reset_index(drop=True, inplace=True)
        self.df_info.reset_index(drop=True, inplace=True)

        self.print(['Add new variables to df_data & df_info: ', self.str_formater(list(dict_add_new_qres.keys())), ' - Completed'], [None, self.clr_blue, None])


        return self.df_data, self.df_info



    def align_ma_values_to_left(self, qre_name: str | list, fillna_val: float = None) -> pd.DataFrame:
        """
        :param qre_name: MA likes 'Q1|8'
        :param fillna_val: fil nan with float value
        :return: df_data
        """
        lst_qre_name = [qre_name] if isinstance(qre_name, str) else qre_name

        for qre_item in lst_qre_name:

            qre, max_col = qre_item.rsplit('|', 1)

            lst_qre = [f'{qre}_{i}' for i in range(1, int(max_col) + 1)]

            df_fil = self.df_data.loc[:, lst_qre].copy()
            df_fil = df_fil.T
            df_sort = pd.DataFrame(np.sort(df_fil.values, axis=0), index=df_fil.index, columns=df_fil.columns)
            df_sort = df_sort.T
            self.df_data[lst_qre] = df_sort[lst_qre]

            del df_fil, df_sort

            if fillna_val:
                self.df_data.loc[self.df_data.eval(f"{qre}_1.isnull()"), f'{qre}_1'] = fillna_val

        self.print(['Align MA values to left: ', self.str_formater(lst_qre_name), ' - Completed'], [None, self.clr_blue, None], end='')
        print(end='\r')


        return self.df_data



    def remove_qres(self, lst_col: list) -> (pd.DataFrame, pd.DataFrame):
        """
        :param lst_col: columns to remove
        :return: df_data, df_info
        """
        self.df_data.drop(columns=lst_col, inplace=True)
        self.df_info = self.df_info.loc[self.df_info.eval(f"~var_name.isin({lst_col})"), :].copy()

        self.df_data.reset_index(drop=True, inplace=True)
        self.df_info.reset_index(drop=True, inplace=True)

        self.print(['Remove questions: ', self.str_formater(lst_col), ' - Completed'], [None, self.clr_blue, None])

        return self.df_data, self.df_info



    def merge_qres(self, *, lst_merge: list, lst_to_merge: list, dk_code: int) -> pd.DataFrame:
        """
        :param lst_merge: output columns
        :param lst_to_merge: input columns
        :param dk_code:
        :return: df_data
        """

        codelist = self.df_info.loc[self.df_info.eval("var_name == @lst_merge[0]"), 'val_lbl'].values.tolist()[0]

        if len(lst_merge) < len(codelist.keys()):
            self.print(f"Merge_qres(error): Length of lst_merge should be greater than or equal length of codelist!!!\nlst_merge = {lst_merge}\ncodelist = {codelist}\nProcessing terminated!!!", self.err)
            exit()


        def merge_row(sr_row: pd.Series, lst_col_name: list, dk: int) -> pd.Series:

            lst_output = sr_row.reset_index(drop=True).drop_duplicates(keep='first').dropna().sort_values().values.tolist()
            output_len = len(lst_col_name)

            if len(lst_output) > 1 and dk in lst_output:
                lst_output.remove(dk)

            if len(lst_output) < output_len:
                lst_output.extend([np.nan] * (output_len - len(lst_output)))

            return pd.Series(data=lst_output, index=lst_col_name)

        self.df_data[lst_merge] = self.df_data[lst_to_merge].apply(merge_row, lst_col_name=lst_merge, dk=dk_code, axis=1)

        self.print(['Merge questions: ', self.str_formater(lst_to_merge), ' - Completed'], [None, self.clr_blue, None])

        return self.df_data



    def convert_percentage(self, lst_qres: list[str], fil_nan: float, is_check_sum: bool) -> (pd.DataFrame, pd.DataFrame):
        """
        :param lst_qres: MA likes 'Q1|8'
        :param fil_nan: fill nan value with float
        :param is_check_sum: check sum for share question (these should be 100%)
        :return: df_data, df_info
        """

        df_check_sum = self.df_data['ID']

        for qre in lst_qres:
            self.print(['Convert percentage', qre], [None, self.clr_blue], sep=': ', end='')
            lst_qre = self.convert_ma_pattern(qre) if '|' in qre else [qre]

            self.df_info.loc[self.df_info.eval("var_name.isin(@lst_qre)"), 'var_type'] = 'NUM'
            self.df_data[lst_qre] = self.df_data[lst_qre].replace('%| ', '', regex=True).astype(float)


            if fil_nan is not None:
                self.df_data[lst_qre] = self.df_data[lst_qre].fillna(fil_nan)

            if is_check_sum:
                df_check_sum = pd.concat([df_check_sum, self.df_data[lst_qre].sum(axis=1)], axis=1)
                df_check_sum.rename(columns={0: f'{qre.rsplit('|', 1)[0]}_Sum'}, inplace=True)

            print(end='\r')


        if is_check_sum:
            df_check_sum = df_check_sum.melt(id_vars=['ID']).query("value != 100")

            if not df_check_sum.empty:
                df_check_sum.to_csv('df_check_sum.csv')
                self.print(f"Please check the percentage of ID: \n{df_check_sum} \n saved with 'df_check_sum.csv'", self.clr_err)

        self.print(['Convert percentage: ', self.str_formater(lst_qres), ' - Completed'], [None, self.clr_blue, None])

        return self.df_data, self.df_info



    @staticmethod
    def convert_ma_pattern(str_ma: str) -> list:
        ma_prefix, ma_suffix = str_ma.rsplit('|', 1)
        return [f'{ma_prefix}_{i}' for i in range(1, int(ma_suffix) + 1)]



    @staticmethod
    def update_append_remove(row, method, lst_val_update) -> pd.Series:

        max_len = row.shape[0]
        lst_val = list()

        match method:

            case 'a':
                lst_val = row.dropna().values.tolist()
                lst_val.extend(lst_val_update)
                lst_val = list(dict.fromkeys(lst_val))

            case 'r':
                update_row = row.replace(lst_val_update, np.nan)
                lst_val = update_row.dropna().values.tolist()


        if len(lst_val) != max_len:
            lst_val = lst_val + [np.nan] * (max_len - len(lst_val))

        update_row = pd.Series(index=row.index, data=lst_val)

        return update_row



    def update_qres_data(self, *, query_fil: str, qre_name: str, lst_val_update: list[int | float], method: str) -> pd.DataFrame:
        """
        :param query_fil:
        :param qre_name: MA likes 'Q1|8'
        :param lst_val_update: list[int | float]
        :param method: 'a' = append, 'r' = remove, 'o' = overlay
        :return: df_data
        """

        # Format 'query_fil'--------------------------------------------------------------------------------------------


        # END Format 'query_fil'----------------------------------------------------------------------------------------

        lst_qre_update = self.convert_ma_pattern(qre_name) if '|' in qre_name else [qre_name]

        match method:

            case 'a' | 'r':
                self.df_data.loc[self.df_data.eval(query_fil), lst_qre_update] = self.df_data.loc[self.df_data.eval(query_fil), lst_qre_update].apply(self.update_append_remove, method=method, lst_val_update=lst_val_update, axis=1)

            case 'o':

                if len(lst_qre_update) != len(lst_val_update):
                    self.print("Length of update columns must equal update values!!!!", self.clr_err)
                    return pd.DataFrame()

                else:
                    self.df_data.loc[self.df_data.eval(query_fil), lst_qre_update] = lst_val_update

            case _:
                self.print(f'Please check param method - {method}', self.clr_err)
                return pd.DataFrame()

        self.print(['Update data of questions: ', self.str_formater(lst_qre_update), ' - Completed'], [None, self.clr_blue, None])

        return self.df_data



    def create_count_ma_ranking(self, *, act: str, lst_qre: list) -> (dict, dict):

        dict_add_new_qres = dict()
        dict_data_new_qres = dict()

        def re_ranking(row, ranking_range: int):

            lst_result = [np.nan] * ranking_range

            df = row.to_frame(name='val')
            df['score'] = list(range(1, df.shape[0] + 1)[::-1])

            for k, v in df.set_index('val', drop=True).to_dict()['score'].items():
                lst_result[k - 1] = v

            return lst_result


        for qre in lst_qre:

            match act.upper():

                case 'MA':
                    lst_col_qre = self.df_info.loc[self.df_info.eval(f"var_name.str.contains(r'^{qre}_[\\d]{{1,2}}$') & var_type.isin(['MA', 'MA_mtr'])"), 'var_name'].values.tolist()

                    if not len(lst_col_qre):
                        self.print(f"{qre} is not {act.upper()} questions!!!", self.clr_err)
                        continue

                    new_num_qre = f'{qre}_Count'
                    dict_add_new_qres.update({new_num_qre: [new_num_qre, 'NUM', {}, np.nan]})
                    dict_data_new_qres.update({new_num_qre: self.df_data[lst_col_qre].count(axis=1).values.tolist()})

                case 'RANKING':
                    lst_col_qre = self.df_info.loc[self.df_info.eval(f"var_name.str.contains(r'^{qre}_Rank[\\d]{{1,2}}$') & var_type.isin(['RANKING'])"), 'var_name'].values.tolist()

                    if not len(lst_col_qre):
                        self.print(f"{qre} is not {act.upper()} questions!!!", self.clr_err)
                        continue

                    dict_codelist = self.df_info.loc[self.df_info.eval("var_name == @lst_col_qre[0]"), 'val_lbl'].values[0]
                    new_num_qre = f'{qre}_ScoreOfAtt'
                    dict_add_new_qres.update(
                        {f"{new_num_qre}_{k}": [f"{k}. {v}", 'NUM', {}, np.nan] for k, v in dict_codelist.items()}
                    )

                    df_ranking: pd.DataFrame = self.df_data[lst_col_qre].apply(re_ranking, ranking_range=len(list(dict_add_new_qres.keys())), axis=1, result_type='expand')
                    df_ranking.columns = list(dict_add_new_qres.keys())
                    dict_data_new_qres = {k: list(v.values()) for k, v in df_ranking.to_dict().items()}

                case _:
                    self.print(f"{act} is not in [MA, RANKING]!!!", self.clr_err)
                    continue

        return dict_add_new_qres, dict_data_new_qres



    def count_ma_choice(self, *, lst_ma_qre: list, dict_replace: dict = None) -> (pd.DataFrame, pd.DataFrame):

        dict_add_new_qres, dict_data_new_qres = self.create_count_ma_ranking(act='MA', lst_qre=lst_ma_qre)

        self.add_qres(dict_add_new_qres)
        self.df_data[list(dict_data_new_qres.keys())] = pd.DataFrame.from_dict(dict_data_new_qres).replace(dict_replace) if dict_replace else pd.DataFrame.from_dict(dict_data_new_qres)

        self.print(['Count MA choice of questions: ', self.str_formater(lst_ma_qre), ' - Completed'], [None, self.clr_blue, None])

        return self.df_data, self.df_info



    def calculate_ranking_score(self, *, lst_ranking_qre: list) -> (pd.DataFrame, pd.DataFrame):

        dict_add_new_qres, dict_data_new_qres = self.create_count_ma_ranking(act='RANKING', lst_qre=lst_ranking_qre)

        self.add_qres(dict_add_new_qres)
        self.df_data[list(dict_data_new_qres.keys())] = pd.DataFrame.from_dict(dict_data_new_qres)

        self.print(['Calculate ranking score of questions: ', self.str_formater(lst_ranking_qre), ' - Completed'], [None, self.clr_blue, None])

        return self.df_data, self.df_info









import pandas as pd
import numpy as np
from pydantic import BaseModel, Field, ConfigDict, PositiveInt, field_validator, model_validator
from typing import List, Dict, Any, Self
from enum import Enum
from .model_data2run import Data2Run, log
from .model_question import *
from .table_horizontal import TableHorizontal
from .table_vertical import TabulationVertical



class CellContent(str, Enum):
    C0 = 'c'
    P0 = 'p'
    P1 = '%'



class SigTestType(str, Enum):
    NON = 'non'
    IND = 'ind'
    REL = 'rel'



class SigTest(BaseModel):
    type: SigTestType = Field(min_length=3, max_length=3, default='non')
    lvl: List[PositiveInt] = Field(min_items=0, max_items=2)
    cols: List[str] = Field(min_items=0, max_items=24)

    @field_validator('lvl')
    @classmethod
    def valcheck_lvl(cls, lvl) -> List[PositiveInt]:

        lst_err = list()
        for i in lvl:
            if not 0 < i < 100:
                lst_err.extend([i])

        if lst_err:
            raise ValueError(f"significant test lvl should in range (0 < i < 100) instead of {lst_err}")

        log.print(f"SigTest validating", log.clr_cyan)
        log.print(f" - Attribute 'lvl'")

        return lvl


    @field_validator('cols')
    @classmethod
    def valcheck_cols(cls, cols) -> List[str]:

        lst_err = list()
        for i in cols:
            x = re.match(r'^[A-Z]$', i)

            if not x:
                lst_err.extend([i])

        if lst_err:
            raise ValueError(f"significant test cols should in range [A-Z] instead of {lst_err}")


        log.print(f" - Attribute 'cols'")

        return cols



class Table(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)

    # INPUT
    tbl_name: str = Field(min_length=1, max_length=30, frozen=True)
    tbl_filter: str = Field(min_length=0, frozen=True)
    tbl_cell_content: List[CellContent] = Field(min_items=1, max_items=3, frozen=True)
    sig_test: SigTest

    is_hide_zero_cats: bool = False
    is_hide_zero_cols: bool = False
    weight_var: str = Field(min_length=0, frozen=True)

    tbl_header: Dict[str, Any]
    tbl_side: List[Question] = Field(min_items=1, examples=[{'name': 'Q1'}, {'name': '$Q2'}])

    data_to_run: Data2Run

    # OUTPUT
    df_horizontal: pd.DataFrame = pd.DataFrame()
    df_vertical: pd.DataFrame = pd.DataFrame()





    @model_validator(mode='after')
    def valcheck_tbl_infor(self) -> Self:

        # # # Check tbl_header -----------------------------------------------------------------------------------------
        lst_err = list()
        df_info = self.data_to_run.df_info
        dict_header_val = self.tbl_header

        for key_hd_grp, val_hd_grp in dict_header_val.items():
            for hd_grp_idx, hd_grp_item in enumerate(val_hd_grp):

                if len(hd_grp_item) == 0:
                    lst_err.extend([f" - Length of {item} @ {key_hd_grp} @ index[{hd_grp_idx}] equals zero"])
                    continue

                for qre in hd_grp_item:

                    qre_name = qre['qre_name']

                    match qre_name:

                        case qre_name if '@' in qre_name:
                            qre.update({'type': 'QUERY'})
                            continue

                        case qre_name if '$' in qre_name:
                            df_qre = df_info.query(f"var_name.str.contains('^{qre_name[1:]}_[0-9]{{1,2}}$')")

                        case _:
                            df_qre = df_info.query(f"var_name == '{qre_name}'")

                    if df_qre.empty:
                        lst_err.extend(
                            [f" - Question {item} @ {key_hd_grp} @ index[{hd_grp_idx}] @ {qre_name} is not found."])
                        continue

                    if qre['cats'] != {}:
                        s_diff = set(qre['cats'].keys()).difference(df_qre.iloc[0, -1].keys())

                        if s_diff.issuperset({'-1'}):
                            s_diff.remove('-1')

                        if len(s_diff) > 0:
                            lst_err.extend([
                                               f" - {item} @ {key_hd_grp} @ index[{hd_grp_idx}] @ {qre_name} @ {s_diff} is not in {df_qre.iloc[0, -1]}"])
                            continue

                    df_qre = df_qre.reset_index(drop=True)
                    dict_update_qre = {
                        'cats': df_qre.at[0, 'val_lbl_unnetted'] if qre['cats'] == {} else qre['cats'],
                        'type': df_qre.at[0, 'var_type'],
                        'lst_col': df_qre['var_name'].values.tolist(),
                    }

                    if len(qre['qre_lbl']) == 0:
                        dict_update_qre.update({'qre_lbl': df_qre.at[0, 'var_lbl']})

                    qre.update(dict_update_qre)

        arr_header_len = np.array([len(i) for i in list(dict_header_val.values())])

        if arr_header_len.min() != arr_header_len.max() != arr_header_len.mean():
            lst_err.extend([f" - Level {item} got an error: {arr_header_len}"])


        if len(lst_err):
            raise AttributeError(f"Invalid dict_header: \n{'\n'.join(lst_err)}")

        log.print(f"Table {self.tbl_name}'s header is checked", log.clr_succ)



        # # # Fill Question Attribute-----------------------------------------------------------------------------------
        for qre in self.tbl_side:

            if '$' in qre.name:
                str_query = f"var_name.str.contains(r'^{qre.name.replace('$', '')}_[0-9]{{1,2}}$')"
            else:
                str_query = f"var_name == '{qre.name}'"

            df_qre = self.data_to_run.df_info.query(str_query)

            if df_qre.empty:
                raise ValueError(f'{qre.name} is not found!!!')

            qre.fill_question(df_qre)

        log.print(f"Table {self.tbl_name}'s tbl_side[question] is filled successfully", log.clr_succ)



        # # # Generate df_horizontal------------------------------------------------------------------------------------
        tbl_hrz = TableHorizontal(obj_table=self)
        tbl_hrz.generate_df_horizontal()
        log.print(f"Table {self.tbl_name}'s df_horizontal is generated successfully", log.clr_succ)



        # # # Generate df_vertical--------------------------------------------------------------------------------------
        tbl_ver = TabulationVertical(obj_table=self)
        tbl_ver.generate_df_vertical()

        here = 1

        log.print(f"Table {self.tbl_name}'s df_vertical is generated successfully", log.clr_succ)












        return self


import pandas as pd
import numpy as np
from pydantic import BaseModel, Field, ConfigDict, model_validator
from typing import List, Self
from .model_side_qre import SideQre



class SideQres(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    qres: List[SideQre] | None = Field(min_items=1, examples=[{'name': 'Q1'}, {'name': '$Q2'}])
    df_info: pd.DataFrame | None

    @model_validator(mode='after')
    def fill_side_qres(self) -> Self:

        if not self.qres:
            return self

        for qre in self.qres:

            if '$' in qre.name:
                str_query = f"var_name.str.contains(r'^{qre.name.replace('$', '')}_[0-9]{{1,2}}$')"
            else:
                str_query = f"var_name == '{qre.name}'"

            df_qre = self.df_info.query(str_query)

            if df_qre.empty:
                raise ValueError(f'{qre.name} is not found!!!')


            if qre.label is None:
                qre.label = df_qre['var_lbl'].values[0]

            elif '{lbl}' in qre.label:
                qre.label = qre.label.replace('{lbl}', df_qre['var_lbl'].values[0])


            qre.type = df_qre['var_type'].values[0]


            if qre.cats is None:
                qre.cats = df_qre['val_lbl'].values[0]


            qre.cols = df_qre['var_name'].values.tolist()


            if (qre.type not in [QreType.SA, QreType.SA_MTR]) and qre.mean_factor is not None:
                raise ValueError(f"Attribute 'mean_factor' only use for SA question!!!")


            if qre.type != QreType.NUM and qre.num_att is not None:
                raise ValueError(f"Attribute 'num_att' only use for NUM question!!!")

        return self














from .ap_data_converter import APDataConverter
from .data_processing import DataProcessing
from .data_transpose import DataTranspose
from .table_generator import DataTableGenerator
from .table_formater import TableFormatter
from .codeframe_reader import CodeframeReader
from .calculate_lsm import LSMCalculation
from .data_analysis import DataAnalysis

from .table_generator_v2 import DataTableGeneratorV2
from .table_formater_v2 import TableFormatterV2

from .Tabulation.tabulation import Tabulation



__all__ = [
    'APDataConverter',
    'DataProcessing',
    'DataTranspose',
    'DataTableGenerator',
    'Tabulation',
    'TableFormatter',
    'CodeframeReader',
    'LSMCalculation',
    'DataAnalysis',
    'DataTableGeneratorV2',
    'TableFormatterV2',
]



import multiprocessing
import xlwings as xw
import re as regex
import time
import datetime
from openpyxl.utils import get_column_letter
from colorama import Fore
from xlwings import sheets, books, Range
from .logging import Logging



class TableFormatterV2(Logging):

    def __init__(self, xlsx_name):
        super().__init__()
        self.xlsx_name = xlsx_name.rsplit('/', 1)[-1] if '/' in xlsx_name else xlsx_name




    def format_workbook(self, *, is_visible: bool = False):

        # is_visible = True

        st = time.time()

        app = xw.App(visible=is_visible)
        app.display_alerts = False

        try:

            wb = xw.Book(self.xlsx_name)

            for ws in wb.sheets:

                ws.activate()

                self.format_worksheet(wb=wb, ws=ws)


            str_output_name = str(self.xlsx_name).replace('.xlsx', f" - {datetime.datetime.now().strftime("%b %d")}.xlsx")

            wb.save(str_output_name)
            wb.close()

            app.display_alerts = True

        except Exception as e:
            raise e

        finally:
            app.quit()

        self.print(f"All sheets formatting have completed in {datetime.timedelta(seconds=time.time() - st)}", self.clr_succ)


    def format_worksheet(self, wb: books, ws: sheets):

        if ws.name == 'Content':
            self.format_ws_content(wb=wb, ws=ws)
        else:
            self.format_ws_table(ws=ws)



    def format_ws_content(self, wb: books, ws: sheets):

        ws.range('B:B').column_width = 50

        lst_ws = wb.sheet_names

        for i, v in enumerate(lst_ws):

            if i == 0:
                continue

            content_cell = ws.range(f'B{i + 1}')
            content_cell.api.Hyperlinks.Add(Anchor=content_cell.api, Address=f"#'{v}'!B3", TextToDisplay=v)

            ws_tbl = wb.sheets[v]
            table_cell = ws_tbl.range('B3')
            table_cell.api.Hyperlinks.Add(Anchor=table_cell.api, Address=f"#'Content'!B{i + 1}", TextToDisplay='Content')

        self.print(['Format sheet', ws.name, 'Completed'], [None, self.clr_blue, None], sep=' | ')



    @staticmethod
    def hex_to_rgb_int(hex_color: str) -> int:
        hex_color = hex_color.lstrip('#')
        rgb = tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))
        return rgb[2] * 65536 + rgb[1] * 256 + rgb[0]  # Convert to Excel RGB integer format



    @staticmethod
    def get_filter_range(visible_cells: Range) -> list:
        return regex.split(r'R|C', visible_cells.GetAddress(ReferenceStyle=False).split(',')[-1].replace(':', ''))[1:]



    def format_ws_table(self, ws: sheets):

        st = time.time()

        ws.range('B:B').column_width = 12
        ws.range('C:C').column_width = 30
        ws.range('F:F').column_width = 40

        first_row = ws.range('A2').end('down').row
        first_column = ws.range('A2').end('right').column

        last_cell = ws.used_range.last_cell
        last_row = last_cell.row
        last_column = last_cell.column

        is_sig_tbl = True

        if ws.range('C4').value is None or 'Weighted' in ws.range('C4').value:
            is_sig_tbl = False

            lst_col_remove = list(range(first_column + 1, last_column + 1, 2))
            lst_col_remove.reverse()

            for icol in lst_col_remove:
                ws.range(f'{get_column_letter(icol)}:{get_column_letter(icol)}').api.Delete()

            # ws.range(f'{first_row - 2}:{first_row - 1}').api.Delete()

            first_row = ws.range('A2').end('down').row
            first_column = ws.range('A2').end('right').column

            last_cell = ws.used_range.last_cell
            last_row = last_cell.row
            last_column = last_cell.column

        rng_content = ws.range(f'{get_column_letter(first_column)}{first_row}:{get_column_letter(last_column)}{last_row}')


        match ws.range('C3').value:
            case 'Cell content: percentage(%)':
                rng_content.api.NumberFormat = '0%'

            case _:
                rng_content.api.NumberFormat = '0'

        idx_col_data = list(dict.fromkeys(ws.range(f'A2:A{last_row}').value))
        idx_col_data.remove(None)


        # BORDERING SIDE QRE--------------------------------------------------------------------------------------------
        for i in idx_col_data:
            ws.range('1:1').api.AutoFilter(Field=1, Criteria1=i)
            filter_range = ws.api.AutoFilter.Range
            visible_cells = filter_range.SpecialCells(12)
            lst_address = self.get_filter_range(visible_cells)
            self.format_ws_qre(ws=ws, lst_address=lst_address)
            ws.api.AutoFilterMode = False


        rng_all_qre_full = ws.range(f'B{first_row}:F{last_row}')
        rng_all_qre_full.api.Borders(11).Weight = 2

        # FORMAT ATTS: mean, std,...------------------------------------------------------------------------------------
        self.format_ws_qre_atts(ws=ws)


        # BORDERING HEADER----------------------------------------------------------------------------------------------
        dict_header_info = {
            'is_sig_tbl': is_sig_tbl,
            'first_row': 3,
            'first_col': first_column,
            'last_row': first_row - 1,
            'last_col': last_column,
            'color': {
                0: {'fontColor': '#FFFFFF', 'fgColor': '#153D64'},
                1: {'fontColor': '#FFFFFF', 'fgColor': '#215C98'},
                2: {'fontColor': '#000000', 'fgColor': '#4D93D9'},
                3: {'fontColor': '#000000', 'fgColor': '#A6C9EC'},

                4: {'fontColor': '#FFFFFF', 'fgColor': '#196B24'},
                5: {'fontColor': '#000000', 'fgColor': '#47D359'},
                6: {'fontColor': '#000000', 'fgColor': '#83E28E'},
                7: {'fontColor': '#000000', 'fgColor': '#C1F0C8'},

                8: {'fontColor': '#FFFFFF', 'fgColor': '#E97132'},
                9: {'fontColor': '#000000', 'fgColor': '#F1A983'},
                10: {'fontColor': '#000000', 'fgColor': '#F7C7AC'},
                11: {'fontColor': '#000000', 'fgColor': '#FBE2D5'},
            }
        }

        self.format_ws_header(ws=ws, dict_header_info=dict_header_info)

        self.print(['Format sheet', ws.name, f'Completed in {datetime.timedelta(seconds=round(time.time() - st, 0))}'], [None, self.clr_blue, None], sep=' | ')



    def format_ws_qre(self, *, ws: sheets, lst_address: list):

        start_row = int(lst_address[0])
        last_row = int(lst_address[2])

        start_col = int(lst_address[1])
        last_col = int(lst_address[-1])

        rng_base = ws.range(f"{get_column_letter(start_col)}{start_row}:{get_column_letter(last_col)}{start_row}")
        rng_qre = ws.range(f"{get_column_letter(start_col)}{start_row}:{get_column_letter(last_col)}{last_row}")

        self.print(['Format sheet', ws.name, rng_qre.address, f'{rng_base.value[1]}-{rng_base.value[3]}'], [None, self.clr_blue, self.clr_cyan_light, self.clr_magenta_light], sep=' | ', end='')

        rng_base.api.Font.Bold = True
        rng_base.api.Interior.Color = self.hex_to_rgb_int('#DAEEF3')
        rng_base.api.NumberFormat = '0'

        rng_qre.api.Borders.Weight = 2
        rng_qre.api.Borders(11).Weight = 1
        rng_qre.api.Borders(12).Weight = 1

        for icol in [1, 2]:

            rng_qre_name = ws.range(f"{get_column_letter(start_col + icol)}{start_row}:{get_column_letter(start_col + icol)}{last_row}")
            rng_qre_name.api.Merge()
            rng_qre_name.api.HorizontalAlignment = -4108
            rng_qre_name.api.VerticalAlignment = -4108
            rng_qre_name.api.WrapText = True

            if icol == 1:
                rng_qre_name.api.Font.Color = self.hex_to_rgb_int('#FFFFFF')

            rng_qre_name.api.Interior.Color = self.hex_to_rgb_int('#538DD5' if icol == 1 else '#D9E1F2')

        print(end='\r')


    def format_ws_qre_atts(self, *, ws: sheets):

        lst_att = [
            'mean', 'std',
            'min', 'max',
            '25%', '50%', '75%',
            'calculate', 'friedman_pval',
            'combine', 'net',
            'net0', 'net1', 'net2',
        ]

        def format_att_font(*, rng: Range, attribute: str, is_entire_row: bool):

            match attribute:
                case str(x) if 'mean' in x or 'std' in x:

                    rng.Font.Bold = True

                    if is_entire_row:
                        rng.NumberFormat = '0.00'
                    else:
                        rng.Font.Color = self.hex_to_rgb_int('#0070C0')

                case str(x) if 'min' in x or 'max' in x:

                    if is_entire_row:
                        rng.NumberFormat = '0.00'
                        rng.Font.Italic = True

                    else:
                        rng.Font.Bold = True
                        rng.Font.Color = self.hex_to_rgb_int('#00B050')

                case str(x) if '25%' in x or '50%' in x or '75%' in x:

                    if is_entire_row:
                        rng.NumberFormat = '0.00'
                        rng.Font.Italic = True

                    else:
                        rng.Font.Bold = True
                        rng.Font.Color = self.hex_to_rgb_int('#963634')

                case str(x) if 'calculate' in x or 'friedman_pval' in x:

                    if is_entire_row:
                        rng.NumberFormat = '0.00'
                        rng.Font.Italic = True

                    else:
                        rng.Font.Bold = True
                        rng.Font.Color = self.hex_to_rgb_int('#E26B0A')

                case str(x) if 'combine' in x:

                    if is_entire_row:
                        rng.Font.Bold = True

                    else:
                        rng.Font.Color = self.hex_to_rgb_int('#000000')


                case str(x) if 'net' in x:

                    if is_entire_row:
                        rng.Font.Bold = True

                    else:

                        match attribute:
                            case str(x) if 'net0' in x:
                                rng.Interior.Color = self.hex_to_rgb_int('#A02B93')
                                rng.Font.Color = self.hex_to_rgb_int('#FFFFFF')

                            case str(x) if 'net1' in x:
                                rng.Interior.Color = self.hex_to_rgb_int('#F2CEEF')

                            case str(x) if 'net2' in x:
                                rng.Interior.Color = self.hex_to_rgb_int('#B5E6A2')

                            case _:
                                rng.Font.Underline = 2  # Single underline
                                rng.Font.Color = self.hex_to_rgb_int('#000000')


        for att in lst_att:

            # FORMAT ENTIRE ROWS----------------------------------------------------------------------------------------
            ws.range(f'1:1').api.AutoFilter(Field=5, Criteria1=f'*{att}*')
            filter_range = ws.api.AutoFilter.Range
            visible_cells = filter_range.SpecialCells(12)
            format_att_font(rng=visible_cells, attribute=att, is_entire_row=True)
            ws.api.AutoFilterMode = False

            # FORMAT VAR_LABEL ONLY-------------------------------------------------------------------------------------
            ws.range(f'E1:F1').api.AutoFilter(Field=1, Criteria1=f'*{att}*')
            filter_range = ws.api.AutoFilter.Range
            visible_cells = filter_range.SpecialCells(12)
            format_att_font(rng=visible_cells, attribute=att, is_entire_row=False)
            ws.api.AutoFilterMode = False



    def format_ws_header(self, *, ws: sheets, dict_header_info: dict):

        is_sig_tbl, first_row, first_col, last_row, last_col, dict_color = dict_header_info.values()

        first_col_letter = get_column_letter(first_col)
        last_col_letter = get_column_letter(last_col)

        rng_full_header = ws.range(f"{first_col_letter}{first_row}:{last_col_letter}{last_row}")
        rng_full_header.api.Borders.Weight = 2

        for i, v in enumerate(range(first_row, last_row + 1)):
            rng_header_row = ws.range(f"{first_col_letter}{v}:{last_col_letter}{v}")
            rng_header_row.api.Interior.Color = self.hex_to_rgb_int(dict_color[i]['fgColor'])
            rng_header_row.api.Font.Color = self.hex_to_rgb_int(dict_color[i]['fontColor'])


        last_usage_row = ws.used_range.last_cell.row
        lst_format_row = list(range(first_row, last_row - 2))
        lst_format_row.reverse()
        lst_merge_range = list()  # [start_col, start_row, end_col, end_row, bold_num]

        bold_num = 2 if not is_sig_tbl else 3

        if is_sig_tbl:
            for icol in range(first_col, last_col, 2):
                lst_merge_range.append([icol, last_row, icol + 1, last_row, 0])
                lst_merge_range.append([icol, last_row - 2, icol + 1, last_row - 2, 0])


        for irow in lst_format_row:

            start_merge_col = first_col

            for icol in range(first_col + 1, last_col + 2):

                col_letter = get_column_letter(icol)
                prev_col_letter = get_column_letter(icol - 1)

                sig_char_cell = ws.range(f"{col_letter}{last_row}")

                prev_cell = ws.range(f"{prev_col_letter}{irow}")
                curr_cell = ws.range(f"{col_letter}{irow}")

                if irow == first_row:
                    upper_row_prev_cell = ws.range(f"{prev_col_letter}{irow}")
                    Upper_row_curr_cell = ws.range(f"{col_letter}{irow}")
                else:
                    upper_row_prev_cell = ws.range(f"{prev_col_letter}{irow - 1}")
                    Upper_row_curr_cell = ws.range(f"{col_letter}{irow - 1}")

                if (sig_char_cell.value == 'A' or icol == last_col + 1) and (prev_cell.value != curr_cell.value or upper_row_prev_cell.value != Upper_row_curr_cell.value):
                    lst_merge_range.append([start_merge_col, irow, icol - 1, irow, bold_num])
                    start_merge_col = icol



        def merge_by_range(lst_range: list):

            merge_range = ws.range(f"{get_column_letter(lst_range[0])}{lst_range[1]}:{get_column_letter(lst_range[2])}{lst_range[3]}")
            merge_range.api.Merge()
            merge_range.api.HorizontalAlignment = -4108  # Center alignment (xlCenter)
            merge_range.api.VerticalAlignment = -4108  # Center alignment (xlCenter)
            merge_range.api.Borders.Weight = 2

            header_x_side_range = ws.range(f"{get_column_letter(lst_range[0])}{lst_range[1]}:{get_column_letter(lst_range[2])}{last_usage_row}")

            if is_sig_tbl:
                if lst_range[-1] > 0:
                    header_x_side_range.api.Borders(7).Weight = lst_range[-1]
                    header_x_side_range.api.Borders(10).Weight = lst_range[-1]
                else:
                    header_x_side_range.api.Borders(7).Weight = 2
                    header_x_side_range.api.Borders(10).Weight = 2
                    header_x_side_range.api.Borders(11).Weight = 1

                    sig_col_range = ws.range(f"{get_column_letter(lst_range[2])}{lst_range[1]}:{get_column_letter(lst_range[2])}{last_usage_row}")
                    sig_col_range.api.Font.Bold = True
                    sig_col_range.api.Font.Color = self.hex_to_rgb_int('#FF0000')

            else:
                header_x_side_range.api.Borders(7).Weight = lst_range[-1]
                header_x_side_range.api.Borders(10).Weight = lst_range[-1]


        for item in lst_merge_range:
            merge_by_range(item)


        rng_full_header.api.Font.Bold = True
        rng_full_header.api.WrapText = True
        rng_full_header.api.HorizontalAlignment = -4108  # Center alignment (xlCenter)
        rng_full_header.api.VerticalAlignment = -4108  # Center alignment (xlCenter)


        # Freeze panes at the selected range

        freeze_range = ws.range(f'{get_column_letter(first_col)}{last_row + 1}')
        freeze_range.api.Application.ActiveWindow.FreezePanes = False
        freeze_range.api.Select()
        freeze_range.api.Application.ActiveWindow.FreezePanes = True


        if is_sig_tbl:
            ws.range(f'{last_row - 1}:{last_row - 1}').api.Delete()
        else:
            ws.range(f'{last_row - 1}:{last_row}').api.Delete()

        ws.range('D:E').api.EntireColumn.Hidden = True
        ws.range('1:2').api.EntireRow.Hidden = True
        ws.range('A:A').api.Delete()


































import pandas as pd
import numpy as np
import pingouin as pg
from pptx import Presentation
from pptx.chart.data import XyChartData
from pptx.enum.chart import XL_CHART_TYPE, XL_LABEL_POSITION
from pptx.util import Inches
# from sklearn.linear_model import LinearRegression, LogisticRegression




class DataAnalysis:
    def __init__(self, *, df_data: pd.DataFrame, df_info: pd.DataFrame):

        self.df_data = df_data
        self.df_info = df_info



    def penalty_analysis(self, *, dict_define_pen: dict, output_name: str):

        df_pen = pd.DataFrame(
            columns=['Section', 'Qre', 'Label', 'Ma_SP_Lbl', 'GroupCode', 'GroupCode_Pct', 'GroupCode_x_OL_Mean', 'JAR_x_OL_Mean', 'Penalty_Score', 'Pull_Down_Index'],
            data=[]
        )

        df_info = self.df_info.copy()

        for k_sec, v_sec in dict_define_pen.items():
            print(f'Processing penalty analysis - {k_sec}')

            df_data = self.df_data.query(v_sec.get('query')).copy() if v_sec.get('query') else self.df_data.copy()

            for k_sp, v_sp in df_info.loc[df_info.eval(f"var_name == '{v_sec['prod_pre']}'"), 'val_lbl'].values[0].items():

                df_fil = df_data.query(f"{v_sec['prod_pre']}.isin([{k_sp}])")

                for k_jar, v_jar in v_sec['jar_qres'].items():

                    jar_ol_mean = df_fil.loc[df_fil.eval(f"{k_jar}.isin({v_jar['jar']['code']})"), v_sec['ol_qre']].mean()

                    for grp in ['b2b', 't2b']:
                        grp_count = df_fil.loc[df_fil.eval(f"{k_jar}.isin({v_jar[grp]['code']})"), k_jar].count()
                        grp_base = df_fil.loc[df_fil.eval(f"{k_jar}.notnull()"), k_jar].count()

                        if not grp_base:
                            continue


                        grp_pct = grp_count / grp_base
                        grp_ol_mean = df_fil.loc[df_fil.eval(f"{k_jar}.isin({v_jar[grp]['code']})"), v_sec['ol_qre']].mean()
                        pen_score = jar_ol_mean - grp_ol_mean

                        dict_pen_data_row = {
                            'Section': k_sec,
                            'Qre': k_jar,
                            'Label': v_jar['label'],
                            'Ma_SP_Lbl': v_sp,
                            'GroupCode': v_jar[grp]['label'],
                            'GroupCode_Pct': grp_pct,
                            'GroupCode_x_OL_Mean': grp_ol_mean,
                            'JAR_x_OL_Mean': jar_ol_mean,
                            'Penalty_Score': pen_score,
                            'Pull_Down_Index': grp_pct * pen_score,
                        }

                        if df_pen.empty:
                            df_pen = pd.DataFrame(columns=list(dict_pen_data_row.keys()), data=[dict_pen_data_row.values()])
                        else:
                            df_pen = pd.concat([df_pen, pd.DataFrame(columns=list(dict_pen_data_row.keys()), data=[dict_pen_data_row.values()])], axis=0, ignore_index=True)

        with pd.ExcelWriter(f'{output_name}.xlsx', engine='openpyxl') as writer:
            df_pen.to_excel(writer, sheet_name=f'Penalty_Analysis')



        # HERE
        # # --------------------------------------------------------------------------------------------------------------
        # # create presentation with 1 slide ------
        # prs = Presentation()
        # slide = prs.slides.add_slide(prs.slide_layouts[5])
        #
        # # Export ppt chart after penalty score table
        # chart_data = XyChartData()
        #
        # series_1 = chart_data.add_series('Model 1')
        # series_1.add_data_point(0.7, 2.7)
        #
        # series_2 = chart_data.add_series('Model 2')
        # series_2.add_data_point(1.3, 3.7)
        #
        # series_2 = chart_data.add_series('Model 3')
        # series_2.add_data_point(0.6, 1.3)
        #
        # x, y, cx, cy = Inches(0.5), Inches(2), Inches(9), Inches(4.5)
        #
        # chart = slide.shapes.add_chart(XL_CHART_TYPE.XY_SCATTER, x, y, cx, cy, chart_data)
        #
        # prs.save('chart-01-test.pptx')



    def linear_regression(self, *, dict_define_linear: dict, output_name: str):
        """
        :param dict_define_linear: dict like
        {
            'lnr1': {
                'str_query': '',
                'dependent_vars': ['Q1'],
                'explanatory_vars': ['Q4', 'Q5', 'Q9', 'Q6', 'Q10'],
            },
            'lnr2': {
                'str_query': '',
                'dependent_vars': ['Q1'],
                'explanatory_vars': ['Q4', 'Q5', 'Q9', 'Q10'],
            },
        }
        :param output_name: *.xlsx
        :return: NONE
        """

        # Single: y = b + a*x
        # Multiple: y = b + a1*x1 + a2*x2 + ... + an*xn

        # # HERE: Going to replace with sklearn func
        # regr = LinearRegression()
        #
        # for k_lnr, v_lnr in dict_define_linear.items():
        #     df_data = self.df_data.query(v_lnr['str_query']).copy() if v_lnr['str_query'] else self.df_data.copy()
        #
        #     x = df_data[v_lnr['dependent_vars']]
        #     y = df_data[v_lnr['explanatory_vars']]
        #
        #     regr.fit(x, y)
        #     print(regr.coef_)

        with pd.ExcelWriter(output_name, engine='openpyxl') as writer:
            for k_ln, v_ln in dict_define_linear.items():
                print(f'Processing linear regression - {k_ln}')

                df_data = self.df_data.query(v_ln['str_query']).copy() if v_ln['str_query'] else self.df_data.copy()

                # if have many dependent_vars, have to calculate mean of its
                df_data.loc[:, 'dep_var'] = df_data.loc[:, v_ln['dependent_vars']].mean(axis=1)

                ln_reg = pg.linear_regression(df_data.loc[:, v_ln['explanatory_vars']], df_data['dep_var'])
                ln_reg.to_excel(writer, sheet_name=k_ln)




    def correlation(self, *, dict_define_corr: dict, output_name: str):
        """
        :param dict_define_corr:
        :param output_name:
        :return: NONE
        """

        with pd.ExcelWriter(f'{output_name}', engine='openpyxl') as writer:
            for key, var in dict_define_corr.items():
                print(f'Processing correlation - {key}')

                df_data = self.df_data.query(var['str_query']).copy() if var['str_query'] else self.df_data.copy()

                # if have many dependent_vars, have to calculate mean of its
                df_data.loc[:, 'dep_var'] = df_data.loc[:, var['dependent_vars']].mean(axis=1)

                x = df_data['dep_var']
                df_corr = pd.DataFrame()


                for i, v in enumerate(var['explanatory_vars']):


                    corr = pg.corr(x, df_data[v])

                    corr['method'] = corr.index
                    corr['x'] = '|'.join(var['dependent_vars'])
                    corr['y'] = v
                    corr.index = [f'correlation {i + 1}']
                    corr = corr[['x', 'y'] + list(corr.columns)[:-2]]

                    df_corr = pd.concat([df_corr, corr])

                df_corr.to_excel(writer, sheet_name=key)






    # MORE ANALYSIS HERE










import pandas as pd
import numpy as np
from ..logging import Logging
from .model_table import Table




class Tabulation(Logging):


    def __init__(self, *, tbl_file_name: str, grp_tbl_info: dict[dict]):
        """
        :param tbl_file_name: output xlsx file name
        :param grp_tbl_info: all tables information for processing
        """

        super().__init__()

        # --------------------------------------------------------------------------------------------------------------
        self.tbl_file_name = tbl_file_name.rsplit('/', 1)[-1] if '/' in tbl_file_name else tbl_file_name

        try:
            with open(self.tbl_file_name):
                self.print(f'File: "{self.tbl_file_name}" is accessed >>> Completed', self.clr_succ)

        except PermissionError:
            raise PermissionError(f'Permission Error when access file: "{self.tbl_file_name}" Processing terminated.')

        except FileNotFoundError:
            pass


        # --------------------------------------------------------------------------------------------------------------
        self.grp_tbl_info = grp_tbl_info
        self.lst_model_tables: list[TableInformation] = list()


        # on hold
        # TableFormatter.__init__(self, self.tbl_file_name)

        # is developing
        # TabulationVertical.__init__(self, dict_tbl_info)







    def tabulate_tables(self, *, lst_running_tables_group: list[str]):

        s_err_tbl = set(lst_running_tables_group).difference(set(self.grp_tbl_info.keys()))

        if s_err_tbl:
            raise ValueError(f"Undefined tables: {s_err_tbl}")


        for grp_tbl_name, tbl_info in self.grp_tbl_info.items():

            if grp_tbl_name not in lst_running_tables_group:
                self.print(f"Skip group tables {grp_tbl_name}", self.clr_warn)
                continue

            self.print(f"Process group tables {grp_tbl_name}", self.clr_magenta)

            dict_data_to_run = dict(tbl_info['data_to_run'])

            for tbl_name, tbl_format in tbl_info['tables_format'].items():

                if tbl_name not in tbl_info['tables_to_run']:
                    self.print(f"Skip {tbl_name}", self.clr_warn)
                    continue
                
                self.print(f"Process {tbl_name}", self.clr_blue)

                tbl_format.update({'data_to_run': dict_data_to_run})
                obj_table = Table.from_orm(tbl_format)

                # For checking
                obj_table.df_horizontal.to_csv(f'tbl_format.df_horizontal_{tbl_name}.csv')


                self.lst_model_tables.append(obj_table)

                dict_data_to_run.update({'is_validated': True})








        del self.grp_tbl_info





    # Note:
    #     - fix sig test, do not sig total with other code
    #     - fix sort method


